# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2025 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple script that connects to multiple Crazyflies in sequence and uploads
the lighthouse configuration file. The Crazyflies that successfully received
the file are powered off, while the ones that didn't get it remain powered on.
This could be really helpful if you're dealing with a big swarm of Crazyflies.

Make sure that each Crazyflie has a lighthouse deck attached.
"""
import os
import sys
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.localization import LighthouseConfigWriter
from cflib.utils import uri_helper
from cflib.utils.power_switch import PowerSwitch

file_path = 'lighthouse.yaml'  # Add the path to your .yaml file

# Modify the list of Crazyflies according to your setup
uris = [
    'radio://0/80/2M/E7E7E7E7E7',
    'radio://0/80/2M/E7E7E7E7E8',
    'radio://0/80/2M/E7E7E7E7E9',
    'radio://0/80/2M/E7E7E7E7EA',
    'radio://0/80/2M/E7E7E7E7EB',
    'radio://0/80/2M/E7E7E7E7EC',
]


def write_one(file_name, scf: SyncCrazyflie):
    print(f'Writing to \033[92m{uri}\033[97m...', end='', flush=True)
    writer = LighthouseConfigWriter(scf.cf)
    writer.write_and_store_config_from_file(None, file_name)
    print('Success!')
    time.sleep(1)


if __name__ == '__main__':
    if not os.path.exists(file_path):
        print('Error: file not found!')
        sys.exit(1)

    print(f'Using file {file_path}')

    cflib.crtp.init_drivers()

    for uri in uris:
        try:
            Drone = uri_helper.uri_from_env(default=uri)
            with SyncCrazyflie(Drone, cf=Crazyflie(rw_cache='./cache')) as scf:
                print(f'\033[92m{uri} \033[97mFully connected ', end='', flush=True)
                while scf.is_params_updated() is False:
                    print('.', end='', flush=True)
                    time.sleep(0.1)
                print(f'{scf.is_params_updated()}')
                time.sleep(0.5)
                write_one(file_path, scf)
                ps = PowerSwitch(Drone)
                ps.platform_power_down()
                time.sleep(2)

        except (Exception):
            print(f'Couldnt connect to \033[91m{uri}\033[97m')
            continue



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2020 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Data channel to communicate with an application running in the Crazyflie
"""
import logging

import cflib.crazyflie.platformservice
from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort
from cflib.utils.callbacks import Caller
# from . import Crazyflie

__author__ = 'Bitcraze AB'
__all__ = ['Appchannel']

logger = logging.getLogger(__name__)


class Appchannel:
    def __init__(self, crazyflie):
        self._cf = crazyflie

        self.packet_received = Caller()

        self._cf.add_port_callback(CRTPPort.PLATFORM,
                                   self._incoming)

    def send_packet(self, data):
        packet = CRTPPacket()
        packet.port = CRTPPort.PLATFORM
        packet.channel = cflib.crazyflie.platformservice.APP_CHANNEL
        packet.data = data
        self._cf.send_packet(packet)

    def _incoming(self, packet: CRTPPacket):
        if packet.channel == cflib.crazyflie.platformservice.APP_CHANNEL:
            self.packet_received.call(packet.data)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 - 2020 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
import struct

from .memory_element import MemoryElement
from cflib.utils.callbacks import Syncer

logger = logging.getLogger(__name__)


class MultirangerMemory(MemoryElement):
    """Memory interface for reading the multiranger values"""

    def __init__(self, id, type, size, mem_handler):
        super(MultirangerMemory, self).__init__(id=id, type=type, size=size,
                                                mem_handler=mem_handler)
        self._read_finished_cb = None

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id and self._read_finished_cb:
            unpacked_data = struct.unpack('<'+'H'*int(len(data) / 2), data)
            zone_matrix = []
            for i in range(8):
                zone_matrix.append(unpacked_data[i*8:i*8+8])

            self._read_finished_cb(addr, zone_matrix)

    def read_data(self, read_finished_cb):
        """Write the saved LED-ring data to the Crazyflie"""
        self._read_finished_cb = read_finished_cb
        self.mem_handler.read(self, 0, 128)

    def read_data_sync(self):
        """Write the saved LED-ring data to the Crazyflie"""
        syncer = Syncer()
        self.read_data(syncer.success_cb)
        syncer.wait()
        if syncer.is_success:
            return syncer.success_args[1]
        else:
            return None

    def read_failed(self, mem, addr):
        if mem.id == self.id:
            logger.debug('Read failed')

    def disconnect(self):
        self._read_finished_cb = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Crazyradio CRTP link driver.

This driver is used to communicate with the Crazyflie using the Crazyradio
USB dongle.
"""
import array
import binascii
import logging
import queue
import re
import struct
import threading
import time
from enum import Enum
from queue import Queue
from threading import Semaphore
from threading import Thread
from typing import Any
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from urllib.parse import parse_qs
from urllib.parse import urlparse

import cflib.drivers.crazyradio as crazyradio
from .crtpstack import CRTPPacket
from .exceptions import WrongUriType
from cflib.crtp.crtpdriver import CRTPDriver
from cflib.crtp.radio_link_statistics import RadioLinkStatistics
from cflib.drivers.crazyradio import Crazyradio


__author__ = 'Bitcraze AB'
__all__ = ['RadioDriver']

logger = logging.getLogger(__name__)

_nr_of_retries = 100
_nr_of_arc_retries = 3

DEFAULT_ADDR_A = [0xe7, 0xe7, 0xe7, 0xe7, 0xe7]
DEFAULT_ADDR = 0xE7E7E7E7E7


class _RadioCommands(Enum):
    STOP = 0
    SEND_PACKET = 1
    SET_ARC = 2
    SCAN_SELECTED = 3
    SCAN_CHANNELS = 4


class _SharedRadioInstance():
    def __init__(self, instance_id: int,
                 cmd_queue: 'Queue[Tuple[int, _RadioCommands, Any]]',
                 rsp_queue: Queue,
                 version: float):
        self._instance_id = instance_id
        self._cmd_queue = cmd_queue
        self._rsp_queue = rsp_queue

        self._channel = 2
        self._address = [0xe7]*5
        self._datarate = crazyradio.Crazyradio.DR_2MPS

        self._opened = True

        self.version = version

    def set_channel(self, channel: int):
        self._channel = channel

    def set_address(self, address):
        self._address = address

    def set_data_rate(self, dr):
        self._datarate = dr

    def send_packet(self, data: List[int]) -> crazyradio._radio_ack:
        assert (self._opened)
        self._cmd_queue.put((self._instance_id,
                             _RadioCommands.SEND_PACKET,
                             (self._channel,
                              self._address,
                              self._datarate,
                              data)))
        ack = self._rsp_queue.get()  # type: crazyradio._radio_ack
        return ack

    def set_arc(self, arc):
        assert (self._opened)
        self._cmd_queue.put((self._instance_id,
                             _RadioCommands.SET_ARC,
                             arc))

    def scan_selected(self, selected, packet):
        assert (self._opened)
        self._cmd_queue.put((self._instance_id,
                             _RadioCommands.SCAN_SELECTED,
                             (self._datarate, self._address,
                              selected, packet)))
        return self._rsp_queue.get()

    def scan_channels(self, start: int, stop: int, packet: Iterable[int]):
        assert (self._opened)
        self._cmd_queue.put((self._instance_id,
                             _RadioCommands.SCAN_CHANNELS,
                             (self._datarate, self._address,
                              start, stop, packet)))
        return self._rsp_queue.get()

    def close(self):
        assert (self._opened)
        self._cmd_queue.put((self._instance_id, _RadioCommands.STOP, None))
        self._opened = False


class _SharedRadio(Thread):
    def __init__(self, devid: int):
        Thread.__init__(self)
        self._radio = Crazyradio(devid=devid)
        self._devid = devid
        self.version = self._radio.version

        self.name = 'Shared Radio'

        self._cmd_queue = Queue()  # type: Queue[Tuple[int, _RadioCommands, Any]]  # noqa
        self._rsp_queues = {}  # type: Dict[int, Queue[Any]]
        self._next_instance_id = 0

        self._lock = Semaphore(1)

        self.daemon = True

        self.start()

    def open_instance(self) -> _SharedRadioInstance:
        rsp_queue = Queue()
        with self._lock:
            instance_id = self._next_instance_id
            self._rsp_queues[instance_id] = rsp_queue
            self._next_instance_id += 1

            if self._radio is None:
                self._radio = Crazyradio(devid=self._devid)

        return _SharedRadioInstance(instance_id,
                                    self._cmd_queue,
                                    rsp_queue,
                                    self.version)

    def run(self):
        while True:
            command = self._cmd_queue.get()

            if command[1] == _RadioCommands.STOP:
                with self._lock:
                    del self._rsp_queues[command[0]]
                    if len(self._rsp_queues) == 0:
                        self._radio.close()
                        self._radio = None
            elif command[1] == _RadioCommands.SEND_PACKET:
                channel, address, datarate, data = command[2]
                self._radio.set_channel(channel)
                self._radio.set_address(address)
                self._radio.set_data_rate(datarate)
                ack = self._radio.send_packet(data)
                self._rsp_queues[command[0]].put(ack)
            elif command[1] == _RadioCommands.SET_ARC:
                self._radio.set_arc(command[2])
            elif command[1] == _RadioCommands.SCAN_SELECTED:
                datarate, address, selected, data = command[2]
                self._radio.set_data_rate(datarate)
                self._radio.set_address(address)
                resp = self._radio.scan_selected(selected, data)
                self._rsp_queues[command[0]].put(resp)
            elif command[1] == _RadioCommands.SCAN_CHANNELS:
                datarate, address, start, stop, packet = command[2]
                self._radio.set_data_rate(datarate)
                self._radio.set_address(address)
                resp = self._radio.scan_channels(start, stop, packet)
                self._rsp_queues[command[0]].put(resp)


class RadioManager:
    _radios = []  # type: List[Union[_SharedRadio, None]]
    _lock = Semaphore(1)

    @staticmethod
    def open(devid: int) -> _SharedRadioInstance:
        with RadioManager._lock:
            if len(RadioManager._radios) <= devid:
                padding = [None] * (devid - len(RadioManager._radios) + 1)
                RadioManager._radios.extend(padding)

            shared_radio = RadioManager._radios[devid]
            if not shared_radio:
                shared_radio = _SharedRadio(devid)
                RadioManager._radios[devid] = shared_radio

            return shared_radio.open_instance()

    @staticmethod
    def remove(devid: int):
        with RadioManager._lock:
            RadioManager._radios[devid] = None


class RadioDriver(CRTPDriver):
    """ Crazyradio link driver """

    def __init__(self):
        """ Create the link driver """
        CRTPDriver.__init__(self)
        self._radio = None
        self.uri = ''
        self.link_error_callback = None
        self.radio_link_statistics_callback = None
        self.in_queue = None
        self.out_queue = None
        self._thread = None
        self.needs_resending = True
        self.rate_limit = None

    def connect(self, uri, radio_link_statistics_callback, link_error_callback):
        """
        Connect the link driver to a specified URI of the format:
        radio://<dongle nbr>/<radio channel>/[250K,1M,2M]

        The callback for radio link statistics can be called at any moment from the
        driver to report back the radio link statistics. The callback from linkError
        will be called when a error occurs with
        an error message.
        """

        devid, channel, datarate, address, rate_limit = self.parse_uri(uri)
        self.uri = uri
        self.rate_limit = rate_limit

        if self._radio is None:
            self._radio = RadioManager.open(devid)
            self._radio.set_channel(channel)
            self._radio.set_data_rate(datarate)
            self._radio.set_address(address)
        else:
            raise Exception('Link already open!')

        if self._radio.version >= 0.4:
            self._radio.set_arc(_nr_of_arc_retries)
        else:
            logger.warning('Radio version <0.4 will be obsoleted soon!')

        # Prepare the inter-thread communication queue
        self.in_queue = queue.Queue()
        # Limited size out queue to avoid "ReadBack" effect
        self.out_queue = queue.Queue(1)

        # Launch the comm thread
        self._thread = _RadioDriverThread(self._radio,
                                          self.in_queue,
                                          self.out_queue,
                                          radio_link_statistics_callback,
                                          link_error_callback,
                                          self,
                                          rate_limit)
        self._thread.start()

        self.link_error_callback = link_error_callback

    @staticmethod
    def parse_uri(uri: str):
        # check if the URI is a radio URI
        if not uri.startswith('radio://'):
            raise WrongUriType('Not a radio URI')

        parsed_uri = urlparse(uri)
        parsed_query = parse_qs(parsed_uri.query)
        parsed_path = parsed_uri.path.strip('/').split('/')

        # Open the USB dongle
        if len(parsed_uri.netloc) < 10 and parsed_uri.netloc.isdigit():
            devid = int(parsed_uri.netloc)
        else:
            try:
                devid = crazyradio.get_serials().index(
                    parsed_uri.netloc.upper())
            except ValueError:
                raise Exception('Cannot find radio with serial {}'.format(
                    parsed_uri.netloc))

        channel = 2
        if len(parsed_path) > 0:
            channel = int(parsed_path[0])

        datarate = Crazyradio.DR_2MPS
        if len(parsed_path) > 1:
            if parsed_path[1] == '250K':
                datarate = Crazyradio.DR_250KPS
            if parsed_path[1] == '1M':
                datarate = Crazyradio.DR_1MPS
            if parsed_path[1] == '2M':
                datarate = Crazyradio.DR_2MPS

        address = DEFAULT_ADDR_A
        if len(parsed_path) > 2:
            # We make sure to pad the address with zeros if we do not have the
            # correct length.
            addr = '{:0>10}'.format(parsed_path[2])
            new_addr = struct.unpack('<BBBBB', binascii.unhexlify(addr))
            address = new_addr

        rate_limit = None
        if 'rate_limit' in parsed_query:
            rate_limit = int(parsed_query['rate_limit'][0])

        return devid, channel, datarate, address, rate_limit

    def receive_packet(self, wait=0):
        """
        Receive a packet though the link. This call is blocking but will
        timeout and return None if a timeout is supplied.
        """
        if wait == 0:
            try:
                return self.in_queue.get(False)
            except queue.Empty:
                return None
        elif wait < 0:
            try:
                return self.in_queue.get(True)
            except queue.Empty:
                return None
        else:
            try:
                return self.in_queue.get(True, wait)
            except queue.Empty:
                return None

    def send_packet(self, pk) -> bool:
        """ Send the packet pk though the link """
        try:
            self.out_queue.put(pk, True, 2)
            return True
        except queue.Full:
            if self.link_error_callback:
                self.link_error_callback('RadioDriver: Could not send packet'
                                         ' to copter')
            return False

    def pause(self):
        self._thread.stop()
        self._thread = None

    def restart(self):
        if self._thread:
            return

        self._thread = _RadioDriverThread(self._radio, self.in_queue,
                                          self.out_queue,
                                          self.radio_link_statistics_callback,
                                          self.link_error_callback,
                                          self,
                                          self.rate_limit)
        self._thread.start()

    def close(self):
        """ Close the link. """
        # Stop the comm thread
        self._thread.stop()

        # Close the USB dongle
        if self._radio:
            self._radio.close()
        self._radio = None

        while not self.out_queue.empty():
            self.out_queue.get()

        # Clear callbacks
        self.link_error_callback = None
        self.radio_link_statistics_callback = None

    def _scan_radio_channels(self, radio: _SharedRadioInstance,
                             start=0, stop=125):
        """ Scan for Crazyflies between the supplied channels. """
        return list(radio.scan_channels(start, stop, (0xff,)))

    def scan_selected(self, links):
        to_scan = ()
        for link in links:
            one_to_scan = {}
            uri_data = re.search('^radio://([0-9]+)((/([0-9]+))'
                                 '(/(250K|1M|2M))?)?',
                                 link)

            one_to_scan['channel'] = int(uri_data.group(4))

            datarate = Crazyradio.DR_2MPS
            if uri_data.group(6) == '250K':
                datarate = Crazyradio.DR_250KPS
            if uri_data.group(6) == '1M':
                datarate = Crazyradio.DR_1MPS
            if uri_data.group(6) == '2M':
                datarate = Crazyradio.DR_2MPS

            one_to_scan['datarate'] = datarate

            to_scan += (one_to_scan,)

        found = self._radio.scan_selected(to_scan, (0xFF, 0xFF, 0xFF))

        ret = ()
        for f in found:
            dr_string = ''
            if f['datarate'] == Crazyradio.DR_2MPS:
                dr_string = '2M'
            if f['datarate'] == Crazyradio.DR_250KPS:
                dr_string = '250K'
            if f['datarate'] == Crazyradio.DR_1MPS:
                dr_string = '1M'

            ret += ('radio://0/{}/{}'.format(f['channel'], dr_string),)

        return ret

    def scan_interface(self, address):
        """ Scan interface for Crazyflies """

        if self._radio is None:
            try:
                self._radio = RadioManager.open(0)
            except Exception as e:
                print(e)
                return []
        try:
            serial = crazyradio.get_serials()
        except Exception as e:
            print(e)
            serial = 'N/A'

        logger.info('v%s dongle with serial %s found', self._radio.version,
                    serial[0])
        found = []

        if address is not None:
            # We make sure to pad the address with zeroes to get correct length
            addr = '{:0>10X}'.format(address)
            new_addr = struct.unpack('<BBBBB', binascii.unhexlify(addr))
            self._radio.set_address(new_addr)

        self._radio.set_arc(1)

        self._radio.set_data_rate(crazyradio.Crazyradio.DR_250KPS)

        if address is None or address == DEFAULT_ADDR:
            found += [['radio://0/{}/250K'.format(c), '']
                      for c in self._scan_radio_channels(self._radio)]
            self._radio.set_data_rate(crazyradio.Crazyradio.DR_1MPS)
            found += [['radio://0/{}/1M'.format(c), '']
                      for c in self._scan_radio_channels(self._radio)]
            self._radio.set_data_rate(crazyradio.Crazyradio.DR_2MPS)
            found += [['radio://0/{}/2M'.format(c), '']
                      for c in self._scan_radio_channels(self._radio)]
        else:
            found += [['radio://0/{}/250K/{:X}'.format(c, address), '']
                      for c in self._scan_radio_channels(self._radio)]
            self._radio.set_data_rate(crazyradio.Crazyradio.DR_1MPS)
            found += [['radio://0/{}/1M/{:X}'.format(c, address), '']
                      for c in self._scan_radio_channels(self._radio)]
            self._radio.set_data_rate(crazyradio.Crazyradio.DR_2MPS)
            found += [['radio://0/{}/2M/{:X}'.format(c, address), '']
                      for c in self._scan_radio_channels(self._radio)]

        self._radio.close()
        self._radio = None

        return found

    def get_status(self):
        try:
            radio = RadioManager.open(0)

            ver = radio.version

            radio.close()

            return 'Crazyradio version {}'.format(ver)
        except Exception:
            return 'Crazyradio not found'

    def get_name(self):
        return 'radio'


# Transmit/receive radio thread
class _RadioDriverThread(threading.Thread):
    """
    Radio link receiver thread used to read data from the
    Crazyradio USB driver. """

    def __init__(self, radio, inQueue, outQueue,
                 radio_link_statistics_callback, link_error_callback, link, rate_limit: Optional[int]):
        """ Create the object """
        threading.Thread.__init__(self, name='RadioDriverThread')
        self._radio = radio
        self._in_queue = inQueue
        self._out_queue = outQueue
        self._sp = False
        self._link_error_callback = link_error_callback
        self._radio_link_statistics = RadioLinkStatistics(radio_link_statistics_callback)
        self._retry_before_disconnect = _nr_of_retries
        self.rate_limit = rate_limit

        self._curr_up = 0
        self._curr_down = 1

        self._has_safelink = False
        self._link = link
        self.daemon = True

    def stop(self):
        """ Stop the thread """
        self._sp = True
        try:
            self.join()
        except Exception:
            pass

    def _send_packet_safe(self, cr, packet):
        """
        Adds 1bit counter to CRTP header to guarantee that no ack (downlink)
        payload are lost and no uplink packet are duplicated.
        The caller should resend packet if not acked (ie. same as with a
        direct call to crazyradio.send_packet)
        """
        # packet = bytearray(packet)
        packet[0] &= 0xF3
        packet[0] |= self._curr_up << 3 | self._curr_down << 2
        resp = cr.send_packet(packet)
        if resp and resp.ack and len(resp.data) and \
           (resp.data[0] & 0x04) == (self._curr_down << 2):
            self._curr_down = 1 - self._curr_down
        if resp and resp.ack:
            self._curr_up = 1 - self._curr_up

        return resp

    def run(self):
        """ Run the receiver thread """
        dataOut = array.array('B', [0xFF])
        waitTime = 0
        emptyCtr = 0
        ackStatus = None

        # Try up to 10 times to enable the safelink mode
        for _ in range(10):
            resp = self._radio.send_packet((0xff, 0x05, 0x01))
            if resp and resp.data and tuple(resp.data) == (
                    0xff, 0x05, 0x01):
                self._has_safelink = True
                self._curr_up = 0
                self._curr_down = 0
                break
        self._link.needs_resending = not self._has_safelink

        while (True):
            if (self._sp):
                break

            try:
                if self._has_safelink:
                    ackStatus = self._send_packet_safe(self._radio, dataOut)
                else:
                    ackStatus = self._radio.send_packet(dataOut)
            except Exception as e:
                import traceback

                self._link_error_callback(
                    'Error communicating with crazy radio ,it has '
                    'probably been unplugged!\nException:%s\n\n%s' % (
                        e, traceback.format_exc()))

            # Analyse the in data packet ...
            if ackStatus is None:
                logger.info('Dongle reported ACK status == None')
                continue

            # If no copter, retry
            if ackStatus.ack is False:
                self._retry_before_disconnect = \
                    self._retry_before_disconnect - 1
                if (self._retry_before_disconnect == 0 and
                        self._link_error_callback is not None):
                    self._link_error_callback('Too many packets lost')
                continue
            self._retry_before_disconnect = _nr_of_retries

            data = ackStatus.data

            # If there is a copter in range, the packet is analysed and the
            # next packet to send is prepared
            # TODO: This does not seem to work since there is always a byte filled in the data even with null packets
            if (len(data) > 0):
                inPacket = CRTPPacket(data[0], list(data[1:]))
                self._in_queue.put(inPacket)
                waitTime = 0
                emptyCtr = 0
            else:
                emptyCtr += 1
                if (emptyCtr > 10):
                    emptyCtr = 10
                    # Relaxation time if the last 10 packet where empty
                    waitTime = 0.01
                else:
                    waitTime = 0

            # If there is a rate limit setup, sleep here to force the rate limit
            if self.rate_limit:
                time.sleep(1.0/self.rate_limit)
                waitTime = 0

            # get the next packet to send of relaxation (wait 10ms)
            outPacket = None
            try:
                outPacket = self._out_queue.get(True, waitTime)
            except queue.Empty:
                outPacket = None

            dataOut = array.array('B')

            if outPacket:
                dataOut.append(outPacket.header)
                for X in outPacket.data:
                    if isinstance(X, int):
                        dataOut.append(X)
                    else:
                        dataOut.append(ord(X))
            else:
                # If no packet to send, send a null packet
                dataOut.append(0xFF)

            self._radio_link_statistics.update(ackStatus, outPacket)


def set_retries_before_disconnect(nr_of_retries):
    global _nr_of_retries
    _nr_of_retries = nr_of_retries


def set_retries(nr_of_arc_retries):
    global _nr_of_arc_retries
    _nr_of_arc_retries = nr_of_arc_retries



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2024 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
This module provides tools for tracking statistics related to the communication
link between the Crazyflie and the lib. Currently, it focuses on tracking latency
but is designed to be extended with additional link statistics in the future.
"""
import struct
import time
from threading import Event
from threading import Thread

import numpy as np

from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort
from cflib.utils.callbacks import Caller

__author__ = 'Bitcraze AB'
__all__ = ['LinkStatistics']

PING_HEADER = 0x0
ECHO_CHANNEL = 0


class LinkStatistics:
    """
    LinkStatistics class manages the collection of various statistics related to the
    communication link between the Crazyflie and the lib.

    This class serves as a high-level manager, initializing and coordinating multiple
    statistics trackers, such as Latency. It allows starting and stopping all
    statistics trackers simultaneously. Future statistics can be added to extend
    the class's functionality.

    Attributes:
        _cf (Crazyflie): A reference to the Crazyflie instance.
        latency (Latency): An instance of the Latency class that tracks latency statistics.
    """

    def __init__(self, crazyflie):
        self._cf = crazyflie

        # Flag to track if the statistics are active
        self._is_active = False

        # Universal statistics
        self.latency = Latency(self._cf)

        # Proxy for latency callback
        self.latency_updated = self.latency.latency_updated

        # Callers for radio link statistics
        self.link_quality_updated = Caller()
        self.uplink_rssi_updated = Caller()
        self.uplink_rate_updated = Caller()
        self.downlink_rate_updated = Caller()
        self.uplink_congestion_updated = Caller()
        self.downlink_congestion_updated = Caller()

    def start(self):
        """
        Start collecting all statistics.
        """
        self._is_active = True
        self.latency.start()

    def stop(self):
        """
        Stop collecting all statistics.
        """
        self._is_active = False
        self.latency.stop()

    def radio_link_statistics_callback(self, radio_link_statistics):
        """
        This callback is called by the RadioLinkStatistics class after it
        processes the data provided by the radio driver.
        """
        if not self._is_active:
            return  # Skip processing if link statistics are stopped

        if 'link_quality' in radio_link_statistics:
            self.link_quality_updated.call(radio_link_statistics['link_quality'])
        if 'uplink_rssi' in radio_link_statistics:
            self.uplink_rssi_updated.call(radio_link_statistics['uplink_rssi'])
        if 'uplink_rate' in radio_link_statistics:
            self.uplink_rate_updated.call(radio_link_statistics['uplink_rate'])
        if 'downlink_rate' in radio_link_statistics:
            self.downlink_rate_updated.call(radio_link_statistics['downlink_rate'])
        if 'uplink_congestion' in radio_link_statistics:
            self.uplink_congestion_updated.call(radio_link_statistics['uplink_congestion'])
        if 'downlink_congestion' in radio_link_statistics:
            self.downlink_congestion_updated.call(radio_link_statistics['downlink_congestion'])


class Latency:
    """
    The Latency class measures and tracks the latency of the communication link
    between the Crazyflie and the lib.

    This class periodically sends ping requests to the Crazyflie and tracks
    the round-trip time (latency). It calculates and stores the 95th percentile
    latency over a rolling window of recent latency measurements.

    Attributes:
        _cf (Crazyflie): A reference to the Crazyflie instance.
        latency (float): The current calculated 95th percentile latency in milliseconds.
        _stop_event (Event): An event object to control the stopping of the ping thread.
        _ping_thread_instance (Thread): Thread instance for sending ping requests at intervals.
    """

    def __init__(self, crazyflie):
        self._cf = crazyflie
        self._cf.add_header_callback(self._ping_response, CRTPPort.LINKCTRL, 0)
        self._stop_event = Event()
        self._ping_thread_instance = None
        self.latency = 0
        self.latency_updated = Caller()

    def start(self):
        """
        Start the latency tracking process.

        This method initiates a background thread that sends ping requests
        at regular intervals to measure and track latency statistics.
        """
        if self._ping_thread_instance is None or not self._ping_thread_instance.is_alive():
            self._stop_event.clear()
            self._ping_thread_instance = Thread(target=self._ping_thread, name='ping_thread')
            self._ping_thread_instance.start()

    def stop(self):
        """
        Stop the latency tracking process.

        This method stops the background thread and ceases sending further
        ping requests, halting latency measurement.
        """
        self._stop_event.set()
        if self._ping_thread_instance is not None:
            self._ping_thread_instance.join()
            self._ping_thread_instance = None

    def _ping_thread(self, interval: float = 0.1) -> None:
        """
        Background thread method that sends a ping to the Crazyflie at regular intervals.

        This method runs in a separate thread and continues to send ping requests
        until the stop event is set.

        Args:
            interval (float): The time (in seconds) to wait between ping requests. Default is 0.1 seconds.
        """
        while not self._stop_event.is_set():
            self.ping()
            time.sleep(interval)

    def ping(self) -> None:
        """
        Send a ping request to the Crazyflie to measure latency.

        A ping packet is sent to the Crazyflie with the current timestamp and a
        header identifier to differentiate it from other echo responses. The latency
        is calculated upon receiving the response.
        """
        ping_packet = CRTPPacket()
        ping_packet.set_header(CRTPPort.LINKCTRL, ECHO_CHANNEL)

        # Pack the current time as the ping timestamp
        current_time = time.time()
        ping_packet.data = struct.pack('<Bd', PING_HEADER, current_time)
        self._cf.send_packet(ping_packet)

    def _ping_response(self, packet):
        """
        Callback method for processing the echo response received from the Crazyflie.

        This method is called when a ping response is received. It checks the header
        to verify that it matches the sent ping header before calculating the latency
        based on the timestamp included in the ping request.

        Args:
            packet (CRTPPacket): The packet received from the Crazyflie containing
            the echo response data.
        """
        received_header, received_timestamp = struct.unpack('<Bd', packet.data)
        if received_header != PING_HEADER:
            return
        self.latency = self._calculate_p95_latency(received_timestamp)
        self.latency_updated.call(self.latency)

    def _calculate_p95_latency(self, timestamp):
        """
        Calculate the 95th percentile latency based on recent ping measurements.

        This method records the round-trip time for a ping response and maintains
        a rolling window of latency values to compute the 95th percentile.

        Args:
            timestamp (float): The timestamp from the sent ping packet to calculate
            the round-trip time.

        Returns:
            float: The updated 95th percentile latency in milliseconds.
        """
        if not hasattr(self, '_latencies'):
            self._latencies = []

        instantaneous_latency = (time.time() - timestamp) * 1000
        self._latencies.append(instantaneous_latency)
        if len(self._latencies) > 100:
            self._latencies.pop(0)
        p95_latency = np.percentile(self._latencies, 95)
        return p95_latency



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2018 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import unittest
from unittest.mock import call
from unittest.mock import MagicMock
from unittest.mock import patch

from cflib.crazyflie import Crazyflie
from cflib.crazyflie import Log
from cflib.crazyflie.log import LogConfig
from cflib.utils.callbacks import Caller
from cflib.utils.multiranger import Multiranger


class MultirangerTest(unittest.TestCase):
    FRONT = 'range.front'
    BACK = 'range.back'
    LEFT = 'range.left'
    RIGHT = 'range.right'
    UP = 'range.up'
    DOWN = 'range.zrange'

    OUT_OF_RANGE = 8000

    def setUp(self):
        self.cf_mock = MagicMock(spec=Crazyflie)

        self.log_config_mock = MagicMock(spec=LogConfig)
        self.log_config_mock.data_received_cb = Caller()

        self.log_mock = MagicMock(spec=Log)
        self.cf_mock.log = self.log_mock

        self.front_data = 2345
        self.back_data = 2345
        self.left_data = 123
        self.right_data = 5432
        self.up_data = 3456
        self.down_data = 1212
        self.log_data = {
            self.FRONT: self.front_data,
            self.BACK: self.back_data,
            self.LEFT: self.left_data,
            self.RIGHT: self.right_data,
            self.UP: self.up_data,
            self.DOWN: self.down_data,
        }

    def test_that_log_configuration_is_added_on_connect(self):
        # Fixture
        with patch('cflib.utils.multiranger.LogConfig',
                   return_value=self.log_config_mock):
            sut = Multiranger(self.cf_mock)

            # Test
            sut.start()

            # Assert
            self.log_mock.add_config.assert_called_once_with(
                self.log_config_mock)

    def test_that_log_configuration_is_correct(self):
        # Fixture
        with patch('cflib.utils.multiranger.LogConfig',
                   return_value=self.log_config_mock):
            sut = Multiranger(self.cf_mock)

            # Test
            sut.start()

            # Assert
            self.log_config_mock.add_variable.assert_has_calls([
                call(self.FRONT),
                call(self.BACK),
                call(self.LEFT),
                call(self.RIGHT),
                call(self.UP),
            ])

    def test_that_log_configuration_is_correct_with_zranger(self):
        # Fixture
        with patch('cflib.utils.multiranger.LogConfig',
                   return_value=self.log_config_mock):
            sut = Multiranger(self.cf_mock, zranger=True)

            # Test
            sut.start()

            # Assert
            self.log_config_mock.add_variable.assert_has_calls([
                call(self.FRONT),
                call(self.BACK),
                call(self.LEFT),
                call(self.RIGHT),
                call(self.UP),
                call(self.DOWN)
            ])

    # def test_that_rate_configuration_is_applied(self):
    #     # Fixture
    #     with patch('cflib.utils.multiranger.LogConfig',
    #                return_value=self.log_config_mock):
    #
    #         # Test
    #         Multiranger(self.cf_mock, rate_ms=123)
    #
    #         # Assert
    #         self.log_config_mock.assert_called_once_with('multiranger', 123)

    def test_that_logging_is_started_on_start(self):
        # Fixture
        with patch('cflib.utils.multiranger.LogConfig',
                   return_value=self.log_config_mock):
            sut = Multiranger(self.cf_mock)

            # Test
            sut.start()

            # Assert
            self.log_config_mock.start.assert_called_once_with()

    def test_that_data_callback_is_set(self):
        # Fixture
        with patch('cflib.utils.multiranger.LogConfig',
                   return_value=self.log_config_mock):
            sut = Multiranger(self.cf_mock)

            # Test
            sut.start()

            # Assert
            self.log_config_mock.start.assert_called_once_with()
            self.assertEqual(1, len(
                self.log_config_mock.data_received_cb.callbacks))

    def test_that_the_log_config_is_deleted_on_stop(self):
        # Fixture
        with patch('cflib.utils.multiranger.LogConfig',
                   return_value=self.log_config_mock):
            sut = Multiranger(self.cf_mock)
            sut.start()

            # Test
            sut.stop()

            # Assert
            self.log_config_mock.delete.assert_called_once_with()

    def test_that_using_context_manager_calls_start_and_stop(self):
        # Fixture
        with patch('cflib.utils.multiranger.LogConfig',
                   return_value=self.log_config_mock):

            with Multiranger(self.cf_mock):
                pass

            # Assert
            self.log_config_mock.start.assert_called_once_with()
            self.log_config_mock.delete.assert_called_once_with()

    def test_that_data_received_from_log_is_available_from_up_getter(self):
        self._validate_distance_from_getter(self.up_data / 1000.0, 'up')

    def test_that_none_is_returned_if_up_ranging_is_off_limit(self):
        # Fixture
        self.log_data[self.UP] = self.OUT_OF_RANGE
        self._validate_distance_from_getter(None, 'up')

    def test_that_data_received_from_log_is_available_from_front_getter(self):
        self._validate_distance_from_getter(self.front_data / 1000.0,
                                            'front')

    def test_that_none_is_returned_if_front_ranging_is_off_limit(self):
        # Fixture
        self.log_data[self.FRONT] = self.OUT_OF_RANGE
        self._validate_distance_from_getter(None, 'front')

    def test_that_data_received_from_log_is_available_from_back_getter(self):
        self._validate_distance_from_getter(self.back_data / 1000.0,
                                            'back')

    def test_that_none_is_returned_if_back_ranging_is_off_limit(self):
        # Fixture
        self.log_data[self.BACK] = self.OUT_OF_RANGE
        self._validate_distance_from_getter(None, 'back')

    def test_that_data_received_from_log_is_available_from_left_getter(self):
        self._validate_distance_from_getter(self.left_data / 1000.0,
                                            'left')

    def test_that_none_is_returned_if_left_ranging_is_off_limit(self):
        # Fixture
        self.log_data[self.LEFT] = self.OUT_OF_RANGE
        self._validate_distance_from_getter(None, 'left')

    def test_that_data_received_from_log_is_available_from_right_getter(self):
        self._validate_distance_from_getter(self.right_data / 1000.0,
                                            'right')

    def test_that_none_is_returned_if_right_ranging_is_off_limit(self):
        # Fixture
        self.log_data[self.RIGHT] = self.OUT_OF_RANGE
        self._validate_distance_from_getter(None, 'right')

    def test_that_data_received_from_log_is_available_from_down_getter(self):
        self._validate_distance_from_getter(self.down_data / 1000.0,
                                            'down', zranger=True)

    def test_that_none_is_returned_if_down_ranging_is_off_limit(self):
        # Fixture
        self.log_data[self.DOWN] = self.OUT_OF_RANGE
        self._validate_distance_from_getter(None, 'down', zranger=True)

    def test_that_none_is_returned_from_down_if_zranger_is_disabled(self):
        # Fixture
        del self.log_data[self.DOWN]
        self._validate_distance_from_getter(None, 'down', zranger=False)

    # Helpers ################################################################

    def _validate_distance_from_getter(self, expected, getter_name,
                                       zranger=False):
        # Fixture
        with patch('cflib.utils.multiranger.LogConfig',
                   return_value=self.log_config_mock):
            sut = Multiranger(self.cf_mock, zranger=zranger)

            timestmp = 1234
            logconf = None

            self.log_config_mock.data_received_cb.call(timestmp, self.log_data,
                                                       logconf)

            # Test
            actual = getattr(sut, getter_name)

            # Assert
            self.assertEqual(expected, actual)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2025 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Script for flying a swarm of 8 Crazyflies performing a coordinated spiral choreography
resembling a Christmas tree outline in 3D space. Each drone takes off to a different
height, flies in spiraling circular layers, and changes radius as it rises and descends,
forming the visual structure of a cone when viewed from outside.

The script is using the high level commanded and has been tested with 3 Crazyradios 2.0
and the Lighthouse positioning system.
"""
import math
import time

import cflib.crtp
from cflib.crazyflie.swarm import CachedCfFactory
from cflib.crazyflie.swarm import Swarm
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie

uri1 = 'radio://0/30/2M/E7E7E7E701'
uri2 = 'radio://0/30/2M/E7E7E7E702'
uri3 = 'radio://0/30/2M/E7E7E7E703'

uri4 = 'radio://1/55/2M/E7E7E7E704'
uri5 = 'radio://1/55/2M/E7E7E7E705'
uri6 = 'radio://1/55/2M/E7E7E7E706'

uri7 = 'radio://2/70/2M/E7E7E7E707'
uri8 = 'radio://2/70/2M/E7E7E7E708'

uris = [
    uri1,
    uri2,
    uri3,
    uri4,
    uri5,
    uri6,
    uri7,
    uri8,
    # Add more URIs if you want more copters in the swarm
]


def arm(scf: SyncCrazyflie):
    scf.cf.platform.send_arming_request(True)
    time.sleep(1.0)


# center of the spiral
x0 = 0
y0 = 0
z0 = 0.5

x_offset = 0.4  # Vertical distance between 2 layers
z_offset = 0.5  # Radius difference between 2 layers

starting_x = {
    uri1: x0 + x_offset,
    uri2: x0 + 2*x_offset,
    uri3: x0 + 3*x_offset,
    uri4: x0 + 4*x_offset,
    uri5: x0 - x_offset,
    uri6: x0 - 2*x_offset,
    uri7: x0 - 3*x_offset,
    uri8: x0 - 4*x_offset,
}

starting_z = {
    uri1: z0 + 3*z_offset,
    uri2: z0 + 2*z_offset,
    uri3: z0 + z_offset,
    uri4: z0,
    uri5: z0 + 3*z_offset,
    uri6: z0 + 2*z_offset,
    uri7: z0 + z_offset,
    uri8: z0,
}

starting_yaw = {
    uri1: math.pi/2,
    uri2: -math.pi/2,
    uri3: math.pi/2,
    uri4: -math.pi/2,
    uri5: -math.pi/2,
    uri6: math.pi/2,
    uri7: -math.pi/2,
    uri8: math.pi/2,

}

rotate_clockwise = {
    uri1: False,
    uri2: True,
    uri3: False,
    uri4: True,
    uri5: False,
    uri6: True,
    uri7: False,
    uri8: True,
}

takeoff_dur = {
    uri: value / 0.4
    for uri, value in starting_z.items()
}


def x_from_z(z):
    cone_width = 4  # m
    cone_height = 5  # m
    """
    Returns the radius of the tree with a given z.
    """
    return cone_width/2 - (cone_width/cone_height) * z


def run_shared_sequence(scf: SyncCrazyflie):
    circle_duration = 5  # Duration of a full-circle
    commander = scf.cf.high_level_commander
    uri = scf._link_uri

    commander.takeoff(starting_z[uri], takeoff_dur[uri])
    time.sleep(max(takeoff_dur.values())+1)

    # Go to the starting position
    commander.go_to(starting_x[uri], y0, starting_z[uri], starting_yaw[uri], 4)
    time.sleep(5)

    # Full circle with ascent=0
    commander.spiral(2*math.pi, abs(starting_x[uri]), abs(starting_x[uri]),
                     ascent=0, duration_s=circle_duration, sideways=False, clockwise=rotate_clockwise[uri])
    time.sleep(circle_duration+1)

    # Half circle with ascent=-0.5m
    commander.spiral(math.pi, abs(starting_x[uri]), x_from_z(starting_z[uri]-0.5*z_offset),
                     ascent=-0.5*z_offset, duration_s=0.5*circle_duration, sideways=False,
                     clockwise=rotate_clockwise[uri])
    time.sleep(0.5*circle_duration+0.5)

    # Full circle with ascent=+1.0m
    commander.spiral(2*math.pi, x_from_z(starting_z[uri]-0.5*z_offset), x_from_z(starting_z[uri]+0.5*z_offset),
                     ascent=z_offset, duration_s=circle_duration, sideways=False,
                     clockwise=rotate_clockwise[uri])
    time.sleep(circle_duration+1)

    # Half circle with ascent=-0.5m
    commander.spiral(math.pi, x_from_z(starting_z[uri]+0.5*z_offset), x_from_z(starting_z[uri]),
                     ascent=-0.5*z_offset, duration_s=0.5*circle_duration, sideways=False,
                     clockwise=rotate_clockwise[uri])
    time.sleep(0.5*circle_duration+1)

    commander.land(0, takeoff_dur[uri])
    time.sleep(max(takeoff_dur.values())+3)


if __name__ == '__main__':
    cflib.crtp.init_drivers()
    factory = CachedCfFactory(rw_cache='./cache')
    with Swarm(uris, factory=factory) as swarm:
        # swarm.reset_estimators()
        time.sleep(0.5)
        print('arming...')
        swarm.parallel_safe(arm)
        print('starting sequence...')
        swarm.parallel_safe(run_shared_sequence)
        time.sleep(1)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2014 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example that scans for available Crazyflies with a certain address and lists them.
"""
import cflib.crtp

# Initiate the low level drivers
cflib.crtp.init_drivers()

print('Scanning interfaces for Crazyflies...')
available = cflib.crtp.scan_interfaces(address=int('E7E7E7E7E7', 16)
                                       )
print('Crazyflies found:')
for i in available:
    print(i[0])



#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2021 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import time

import cflib.crtp  # noqa
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper

# Reads the CFLIB_URI environment variable for URI or uses default
uri = uri_helper.uri_from_env(default='radio://0/30/2M/E7E7E7E7E7')


def console_callback(text: str):
    '''A callback to run when we get console text from Crazyflie'''
    # We do not add newlines to the text received, we get them from the
    # Crazyflie at appropriate places.
    print(text, end='')


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    # Create Crazyflie object, with cache to avoid re-reading param and log TOC
    cf = Crazyflie(rw_cache='./cache')

    # Add a callback to whenever we receive 'console' text from the Crazyflie
    # This is the output from DEBUG_PRINT calls in the firmware.
    #
    # This could also be a Python lambda, something like:
    #   cf.console.receivedChar.add_callback(lambda text: logger.info(text))
    cf.console.receivedChar.add_callback(console_callback)

    # This will connect the Crazyflie with the URI specified above.
    # You might have to restart your Crazyflie in order to get output
    # from console, since not much is written during regular uptime.
    with SyncCrazyflie(uri, cf=cf) as scf:
        print('[host] Connected, use ctrl-c to quit.')

        while True:
            time.sleep(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
USB driver for the Crazyradio USB dongle.
"""
import logging
import os
import platform

import libusb_package
import usb
import usb.core

try:
    if os.environ['CRTP_PCAP_LOG'] is not None:
        from cflib.crtp.pcap import PCAPLog
except KeyError:
    pass

__author__ = 'Bitcraze AB'
__all__ = ['Crazyradio']

logger = logging.getLogger(__name__)

# USB parameters
CRADIO_VID = 0x1915
CRADIO_PID = 0x7777

# Dongle configuration requests
# See https://www.bitcraze.io/documentation/repository/crazyradio-firmware/master/functional-areas/usb_radio_protocol/
#  for documentation
SET_RADIO_CHANNEL = 0x01
SET_RADIO_ADDRESS = 0x02
SET_DATA_RATE = 0x03
SET_RADIO_POWER = 0x04
SET_RADIO_ARD = 0x05
SET_RADIO_ARC = 0x06
ACK_ENABLE = 0x10
SET_CONT_CARRIER = 0x20
SCANN_CHANNELS = 0x21
LAUNCH_BOOTLOADER = 0xFF


def _find_devices(serial=None):
    """
    Returns a list of CrazyRadio devices currently connected to the computer
    """
    ret = []

    if os.name == 'nt':
        import usb.backend.libusb0 as libusb0

        backend = libusb0.get_backend()
    else:
        backend = libusb_package.get_libusb1_backend()

    devices = usb.core.find(idVendor=0x1915, idProduct=0x7777, find_all=1,
                            backend=backend)
    if devices:
        for d in devices:
            if serial is not None and serial == d.serial_number:
                return d
            ret.append(d)

    return ret


def get_serials():
    return tuple(map(lambda d: d.serial_number, _find_devices()))


class _radio_ack:
    ack = False
    powerDet = False
    retry = 0
    data = ()


class Crazyradio:
    """ Used for communication with the Crazyradio USB dongle """
    # configuration constants
    DR_250KPS = 0
    DR_1MPS = 1
    DR_2MPS = 2

    P_M18DBM = 0
    P_M12DBM = 1
    P_M6DBM = 2
    P_0DBM = 3

    def __init__(self, device=None, devid=0, serial=None):
        """ Create object and scan for USB dongle if no device is supplied """

        self.current_channel = None
        self.current_address = None
        self.current_datarate = None
        self.devid = devid

        if device is None:
            try:
                if serial is None:
                    device = _find_devices()[devid]
                else:
                    device = _find_devices(serial)
            except Exception:
                if serial is None:
                    raise Exception('Cannot find a Crazyradio Dongle')
                else:
                    raise Exception('Cannot find Crazyradio {}'.format(serial))

        self.dev = device

        self.dev.set_configuration(1)
        self.handle = self.dev
        self.version = float('{0:x}.{1:x}'.format(
            self.dev.bcdDevice >> 8, self.dev.bcdDevice & 0x0FF))

        if self.version < 0.3:
            raise 'This driver requires Crazyradio firmware V0.3+'

        if self.version < 0.4:
            logger.warning('You should update to Crazyradio firmware V0.4+')

        if platform.system() == 'Linux':
            self.handle.reset()

        self.set_data_rate(self.DR_2MPS)
        self.set_channel(2)
        self.arc = -1
        if self.version >= 0.4:
            self.set_cont_carrier(False)
            self.set_address((0xE7,) * 5)
            self.set_power(self.P_0DBM)
            self.set_arc(3)
            self.set_ard_bytes(32)
            self.set_ack_enable(True)

    def _log_packet(self, receive, devid, address, channel, packet):
        try:
            if os.environ['CRTP_PCAP_LOG'] is not None:
                if len(packet) > 0:
                    logger = PCAPLog.instance()
                    logger.logCRTP(logger.LinkType.RADIO, receive, devid, address, channel, packet)
        except KeyError:
            pass

    def close(self):
        if self.dev:
            usb.util.dispose_resources(self.dev)

        self.handle = None
        self.dev = None

        self.current_channel = None
        self.current_address = None
        self.current_datarate = None

    # Dongle configuration
    def set_channel(self, channel):
        """ Set the radio channel to be used """
        if channel != self.current_channel:
            _send_vendor_setup(self.handle, SET_RADIO_CHANNEL, channel, 0, ())
            self.current_channel = channel

    def set_address(self, address):
        """ Set the radio address to be used"""
        if len(address) != 5:
            raise Exception('Crazyradio: the radio address shall be 5'
                            ' bytes long')
        if address != self.current_address:
            _send_vendor_setup(self.handle, SET_RADIO_ADDRESS, 0, 0, address)
            self.current_address = address

    def set_data_rate(self, datarate):
        """ Set the radio datarate to be used """
        if datarate != self.current_datarate:
            _send_vendor_setup(self.handle, SET_DATA_RATE, datarate, 0, ())
            self.current_datarate = datarate

    def set_power(self, power):
        """ Set the radio power to be used """
        _send_vendor_setup(self.handle, SET_RADIO_POWER, power, 0, ())

    def set_arc(self, arc):
        """ Set the ACK retry count for radio communication """
        _send_vendor_setup(self.handle, SET_RADIO_ARC, arc, 0, ())
        self.arc = arc

    def set_ard_time(self, us):
        """ Set the ACK retry delay for radio communication """
        # Auto Retransmit Delay:
        # 0000 - Wait 250uS
        # 0001 - Wait 500uS
        # 0010 - Wait 750uS
        # ........
        # 1111 - Wait 4000uS

        # Round down, to value representing a multiple of 250uS
        t = int((us / 250) - 1)
        if (t < 0):
            t = 0
        if (t > 0xF):
            t = 0xF
        _send_vendor_setup(self.handle, SET_RADIO_ARD, t, 0, ())

    def set_ard_bytes(self, nbytes):
        _send_vendor_setup(self.handle, SET_RADIO_ARD, 0x80 | nbytes, 0, ())

    def set_cont_carrier(self, active):
        if active:
            _send_vendor_setup(self.handle, SET_CONT_CARRIER, 1, 0, ())
        else:
            _send_vendor_setup(self.handle, SET_CONT_CARRIER, 0, 0, ())

    def set_ack_enable(self, enable):
        if enable:
            _send_vendor_setup(self.handle, ACK_ENABLE, 1, 0, ())
        else:
            _send_vendor_setup(self.handle, ACK_ENABLE, 0, 0, ())

    def _has_fw_scan(self):
        # return self.version >= 0.5
        # FIXME: Mitigation for Crazyradio firmware bug #9
        return False

    def scan_selected(self, selected, packet):
        result = ()
        for s in selected:
            self.set_channel(s['channel'])
            self.set_data_rate(s['datarate'])
            status = self.send_packet(packet)
            if status and status.ack:
                result += (s,)

        return result

    def scan_channels(self, start, stop, packet):
        if self._has_fw_scan():  # Fast firmware-driven scan
            self.current_channel = None
            self.current_address = None
            self.current_datarate = None

            _send_vendor_setup(self.handle, SCANN_CHANNELS, start, stop,
                               packet)
            return tuple(_get_vendor_setup(self.handle, SCANN_CHANNELS,
                                           0, 0, 64))
        else:  # Slow PC-driven scan
            result = tuple()
            for i in range(start, stop + 1):
                self.set_channel(i)
                status = self.send_packet(packet)
                if status and status.ack:
                    result += (i,)
            return result

    # Data transfers
    def send_packet(self, dataOut):
        """ Send a packet and receive the ack from the radio dongle
            The ack contains information about the packet transmission
            and a data payload if the ack packet contained any """
        ackIn = None
        data = None
        try:
            self.handle.write(endpoint=1, data=dataOut, timeout=1000)
            data = self.handle.read(0x81, 64, timeout=1000)

            self._log_packet(
                False,
                self.devid,
                self.current_address,
                self.current_channel,
                dataOut
            )
        except usb.USBError:
            pass

        if data is not None:
            ackIn = _radio_ack()
            if data[0] != 0:
                ackIn.ack = (data[0] & 0x01) != 0
                ackIn.powerDet = (data[0] & 0x02) != 0
                ackIn.retry = data[0] >> 4
                ackIn.data = data[1:]
            else:
                ackIn.retry = self.arc

            if ackIn.ack:
                self._log_packet(
                    True,
                    self.devid,
                    self.current_address,
                    self.current_channel,
                    ackIn.data
                )

        return ackIn


# Private utility functions
def _send_vendor_setup(handle, request, value, index, data):
    handle.ctrl_transfer(usb.TYPE_VENDOR, request, wValue=value,
                         wIndex=index, timeout=1000, data_or_wLength=data)


def _get_vendor_setup(handle, request, value, index, length):
    return handle.ctrl_transfer(usb.TYPE_VENDOR | 0x80, request,
                                wValue=value, wIndex=index, timeout=1000,
                                data_or_wLength=length)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Example of how to read the memory from a deck
"""
import logging
from threading import Event

import cflib.crtp  # noqa
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.mem import MemoryElement
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


class ReadMem:
    def __init__(self, uri):
        self._event = Event()
        self._cf = Crazyflie(rw_cache='./cache')

        with SyncCrazyflie(uri, cf=self._cf) as scf:
            mems = scf.cf.mem.get_mems(MemoryElement.TYPE_DECK_MEMORY)

            count = len(mems)
            if count != 1:
                raise Exception('Unexpected nr of memories found:', count)

            mem = mems[0]
            mem.query_decks(self.query_complete_cb)
            self._event.wait()

            if len(mem.deck_memories.items()) == 0:
                print('No memories to read')

            for id, deck_mem in mem.deck_memories.items():
                print('-----')
                print('Deck id:', id)
                print('Name:', deck_mem.name)
                print('is_started:', deck_mem.is_started)
                print('supports_read:', deck_mem.supports_read)
                print('supports_write:', deck_mem.supports_write)

                if deck_mem.supports_fw_upgrade:
                    print('This deck supports FW upgrades')
                    print(
                        f'  The required FW is: hash={deck_mem.required_hash}, '
                        f'length={deck_mem.required_length}, name={deck_mem.name}')
                    print('  is_fw_upgrade_required:', deck_mem.is_fw_upgrade_required)
                    if (deck_mem.is_bootloader_active):
                        print('  In bootloader mode, ready to be flashed')
                    else:
                        print('  In FW mode')
                    print('')

                if deck_mem.supports_read:
                    print('Reading first 10 bytes of memory')
                    self._event.clear()
                    deck_mem.read(0, 10, self.read_complete, self.read_failed)
                    self._event.wait()

    def query_complete_cb(self, deck_memories):
        self._event.set()

    def read_complete(self, addr, data):
        print(data)
        self._event.set()

    def read_failed(self, addr):
        print('Read failed @ {}'.format(addr))
        self._event.set()


if __name__ == '__main__':
    # URI to the Crazyflie to connect to
    uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    rm = ReadMem(uri)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2014 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example that connects to the first Crazyflie found, ramps up/down
the motors and disconnects.
"""
import logging
import time

import cflib
from cflib.crazyflie import Crazyflie
from cflib.utils import uri_helper

uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

logging.basicConfig(level=logging.ERROR)


class MotorRampExample:
    """Example that connects to a Crazyflie and ramps the motors up/down and
    the disconnects"""

    def __init__(self, link_uri):
        """ Initialize and run the example with the specified link_uri """

        self._cf = Crazyflie(rw_cache='./cache')

        self._cf.connected.add_callback(self._connected)
        self._cf.disconnected.add_callback(self._disconnected)
        self._cf.connection_failed.add_callback(self._connection_failed)
        self._cf.connection_lost.add_callback(self._connection_lost)

        self._cf.open_link(link_uri)

        print('Connecting to %s' % link_uri)

    def _connected(self, link_uri):
        """ This callback is called form the Crazyflie API when a Crazyflie
        has been connected and the TOCs have been downloaded."""

        # Arm the Crazyflie
        self._cf.platform.send_arming_request(True)
        time.sleep(1.0)

    def _connection_failed(self, link_uri, msg):
        """Callback when connection initial connection fails (i.e no Crazyflie
        at the specified address)"""
        print('Connection to %s failed: %s' % (link_uri, msg))

    def _connection_lost(self, link_uri, msg):
        """Callback when disconnected after a connection has been made (i.e
        Crazyflie moves out of range)"""
        print('Connection to %s lost: %s' % (link_uri, msg))

    def _disconnected(self, link_uri):
        """Callback when the Crazyflie is disconnected (called in all cases)"""
        print('Disconnected from %s' % link_uri)

    def ramp_motors(self):
        thrust_mult = 1
        thrust_step = 500
        thrust = 20000
        pitch = 0
        roll = 0
        yawrate = 0

        # Unlock startup thrust protection
        self._cf.commander.send_setpoint(0, 0, 0, 0)

        while thrust >= 20000:
            self._cf.commander.send_setpoint(roll, pitch, yawrate, thrust)
            time.sleep(0.1)
            if thrust >= 25000:
                thrust_mult = -1
            thrust += thrust_step * thrust_mult
        for _ in range(30):
            # Continuously send the zero setpoint until the drone is recognized as landed
            # to prevent the supervisor from intervening due to missing regular setpoints
            self._cf.commander.send_setpoint(0, 0, 0, 0)
            # Sleeping before closing the link makes sure the last
            # packet leaves before the link is closed, since the
            # message queue is not flushed before closing
            time.sleep(0.1)

    def disconnect(self):
        self._cf.close_link()


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    me = MotorRampExample(uri)

    me.ramp_motors()

    me.disconnect()



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2016-2022 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
The synchronous Crazyflie class is a wrapper around the "normal" Crazyflie
class. It handles the asynchronous nature of the Crazyflie API and turns it
into blocking functions. It is useful for simple scripts that performs tasks
as a sequence of events.

Example:
```python
with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
    with PositionHlCommander(scf, default_height=0.5, default_velocity=0.2) as pc:
        # fly onto a landing platform at non-zero height (ex: from floor to desk, etc)
        pc.forward(1.0)
        pc.left(1.0)
```
"""
import logging
from threading import Event

from cflib.crazyflie import Crazyflie

logger = logging.getLogger(__name__)


class SyncCrazyflie:

    def __init__(self, link_uri, cf=None):
        """
        Create a synchronous Crazyflie instance with the specified link_uri

        :param link_uri: The uri to use when connecting to the Crazyflie
        :param cf: Optional Crazyflie instance to use, None by default. If no object is supplied, a Crazyflie instance
         is created. This parameters is useful if you want to use a Crazyflie instance with log/param caching.
        """
        if cf:
            self.cf = cf
        else:
            self.cf = Crazyflie()

        self._link_uri = link_uri
        self._connect_event = None
        self._disconnect_event = None
        self._params_updated_event = Event()
        self._is_link_open = False
        self._error_message = None

    def open_link(self):
        """
        Open a link to a Crazyflie on the underlying Crazyflie instance.

        This function is blocking and will return when the connection is established and TOCs for log and
        parameters have been downloaded or fetched from the cache.

        Note: Parameter values have not been updated when this function returns. See the wait_for_params()
        method.
        """
        if (self.is_link_open()):
            raise Exception('Link already open')

        self._add_callbacks()

        logger.debug('Connecting to %s' % self._link_uri)

        self._connect_event = Event()
        self._params_updated_event.clear()
        self.cf.open_link(self._link_uri)
        self._connect_event.wait()
        self._connect_event = None

        if not self._is_link_open:
            self._remove_callbacks()
            self._params_updated_event.clear()
            self.cf.close_link()
            raise Exception(self._error_message)

    def wait_for_params(self):
        """
        Wait for parameter values to be updated.

        During the connection sequence, parameter values are downloaded after the TOCs have been received. The
        open_link() method will return after the TOCs have been received but before the parameter values
        are downloaded.
        This method will block until the parameter values are received and can be used
        to make sure the connection sequence has terminated. In most cases this is not important, but
        radio bandwidth will be limited while parameters are downloaded due to the communication that is going on.

        Example:
        ```python
        with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
            scf.wait_for_params()
            # At this point the connection sequence is finished
        ```

        """
        self._params_updated_event.wait()

    def __enter__(self):
        self.open_link()
        return self

    def close_link(self):
        if (self.is_link_open()):
            self._disconnect_event = Event()
            self.cf.close_link()
            self._disconnect_event.wait()
            self._disconnect_event = None
            self._params_updated_event.clear()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_link()

    def is_link_open(self):
        return self._is_link_open

    def is_params_updated(self):
        return self._params_updated_event.is_set()

    def _connected(self, link_uri):
        """ This callback is called form the Crazyflie API when a Crazyflie
        has been connected and the TOCs have been downloaded."""
        logger.debug('Connected to %s' % link_uri)
        self._is_link_open = True
        if self._connect_event:
            self._connect_event.set()

    def _connection_failed(self, link_uri, msg):
        """Callback when initial connection fails (i.e no Crazyflie
        at the specified address)"""
        logger.debug('Connection to %s failed: %s' % (link_uri, msg))
        self._is_link_open = False
        self._error_message = msg
        if self._connect_event:
            self._connect_event.set()

    def _disconnected(self, link_uri):
        self._remove_callbacks()
        self._is_link_open = False
        if self._disconnect_event:
            self._disconnect_event.set()

    def _all_params_updated(self, link_uri):
        self._params_updated_event.set()

    def _add_callbacks(self):
        self.cf.connected.add_callback(self._connected)
        self.cf.connection_failed.add_callback(self._connection_failed)
        self.cf.disconnected.add_callback(self._disconnected)
        self.cf.fully_connected.add_callback(self._all_params_updated)

    def _remove_callbacks(self):
        def remove_callback(container, callback):
            try:
                container.remove_callback(callback)
            except ValueError:
                pass

        remove_callback(self.cf.connected, self._connected)
        remove_callback(self.cf.connection_failed, self._connection_failed)
        remove_callback(self.cf.disconnected, self._disconnected)
        remove_callback(self.cf.fully_connected, self._all_params_updated)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Example of how to connect to a Qualisys QTM system and feed the position to a
Crazyflie. It uses the high level commander to upload a trajectory to fly a
figure 8.

Set the uri to the radio settings of the Crazyflie and modify the
rigid_body_name to match the name of the Crazyflie in QTM.

Requires the qualisys optional dependency:
    pip install cflib[qualisys]
"""
import asyncio
import math
import time
import xml.etree.cElementTree as ET
from threading import Thread

import qtm_rt
from scipy.spatial.transform import Rotation

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.mem import MemoryElement
from cflib.crazyflie.mem import Poly4D
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper
from cflib.utils.reset_estimator import reset_estimator

# URI to the Crazyflie to connect to
uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# The name of the rigid body in QTM that represents the Crazyflie
rigid_body_name = 'cf'

# True: send position and orientation; False: send position only
send_full_pose = True

# When using full pose, the estimator can be sensitive to noise in the orientation data when yaw is close to +/- 90
# degrees. If this is a problem, increase orientation_std_dev a bit. The default value in the firmware is 4.5e-3.
orientation_std_dev = 8.0e-3

# The trajectory to fly
# See https://github.com/whoenig/uav_trajectories for a tool to generate
# trajectories

# Duration,x^0,x^1,x^2,x^3,x^4,x^5,x^6,x^7,y^0,y^1,y^2,y^3,y^4,y^5,y^6,y^7,z^0,z^1,z^2,z^3,z^4,z^5,z^6,z^7,yaw^0,yaw^1,yaw^2,yaw^3,yaw^4,yaw^5,yaw^6,yaw^7
figure8 = [
    [1.050000, 0.000000, -0.000000, 0.000000, -0.000000, 0.830443, -0.276140, -0.384219, 0.180493, -0.000000, 0.000000, -0.000000, 0.000000, -1.356107, 0.688430, 0.587426, -0.329106, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000],  # noqa
    [0.710000, 0.396058, 0.918033, 0.128965, -0.773546, 0.339704, 0.034310, -0.026417, -0.030049, -0.445604, -0.684403, 0.888433, 1.493630, -1.361618, -0.139316, 0.158875, 0.095799, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000],  # noqa
    [0.620000, 0.922409, 0.405715, -0.582968, -0.092188, -0.114670, 0.101046, 0.075834, -0.037926, -0.291165, 0.967514, 0.421451, -1.086348, 0.545211, 0.030109, -0.050046, -0.068177, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000],  # noqa
    [0.700000, 0.923174, -0.431533, -0.682975, 0.177173, 0.319468, -0.043852, -0.111269, 0.023166, 0.289869, 0.724722, -0.512011, -0.209623, -0.218710, 0.108797, 0.128756, -0.055461, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000],  # noqa
    [0.560000, 0.405364, -0.834716, 0.158939, 0.288175, -0.373738, -0.054995, 0.036090, 0.078627, 0.450742, -0.385534, -0.954089, 0.128288, 0.442620, 0.055630, -0.060142, -0.076163, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000],  # noqa
    [0.560000, 0.001062, -0.646270, -0.012560, -0.324065, 0.125327, 0.119738, 0.034567, -0.063130, 0.001593, -1.031457, 0.015159, 0.820816, -0.152665, -0.130729, -0.045679, 0.080444, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000],  # noqa
    [0.700000, -0.402804, -0.820508, -0.132914, 0.236278, 0.235164, -0.053551, -0.088687, 0.031253, -0.449354, -0.411507, 0.902946, 0.185335, -0.239125, -0.041696, 0.016857, 0.016709, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000],  # noqa
    [0.620000, -0.921641, -0.464596, 0.661875, 0.286582, -0.228921, -0.051987, 0.004669, 0.038463, -0.292459, 0.777682, 0.565788, -0.432472, -0.060568, -0.082048, -0.009439, 0.041158, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000],  # noqa
    [0.710000, -0.923935, 0.447832, 0.627381, -0.259808, -0.042325, -0.032258, 0.001420, 0.005294, 0.288570, 0.873350, -0.515586, -0.730207, -0.026023, 0.288755, 0.215678, -0.148061, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000],  # noqa
    [1.053185, -0.398611, 0.850510, -0.144007, -0.485368, -0.079781, 0.176330, 0.234482, -0.153567, 0.447039, -0.532729, -0.855023, 0.878509, 0.775168, -0.391051, -0.713519, 0.391628, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000],  # noqa
]


class QtmWrapper(Thread):
    def __init__(self, body_name):
        Thread.__init__(self)

        self.body_name = body_name
        self.on_pose = None
        self.connection = None
        self.qtm_6DoF_labels = []
        self._stay_open = True

        self.start()

    def close(self):
        self._stay_open = False
        self.join()

    def run(self):
        asyncio.run(self._life_cycle())

    async def _life_cycle(self):
        await self._connect()
        while (self._stay_open):
            await asyncio.sleep(1)
        await self._close()

    async def _connect(self):
        qtm_instance = await self._discover()
        host = qtm_instance.host
        print('Connecting to QTM on ' + host)
        self.connection = await qtm_rt.connect(host)

        params = await self.connection.get_parameters(parameters=['6d'])
        xml = ET.fromstring(params)
        self.qtm_6DoF_labels = [label.text.strip() for index, label in enumerate(xml.findall('*/Body/Name'))]

        await self.connection.stream_frames(
            components=['6D'],
            on_packet=self._on_packet)

    async def _discover(self):
        async for qtm_instance in qtm_rt.Discover('0.0.0.0'):
            return qtm_instance

    def _on_packet(self, packet):
        header, bodies = packet.get_6d()

        if bodies is None:
            return

        if self.body_name not in self.qtm_6DoF_labels:
            print('Body ' + self.body_name + ' not found.')
        else:
            index = self.qtm_6DoF_labels.index(self.body_name)
            temp_cf_pos = bodies[index]
            x = temp_cf_pos[0][0] / 1000
            y = temp_cf_pos[0][1] / 1000
            z = temp_cf_pos[0][2] / 1000

            r = temp_cf_pos[1].matrix
            rot = [
                [r[0], r[3], r[6]],
                [r[1], r[4], r[7]],
                [r[2], r[5], r[8]],
            ]

            if self.on_pose:
                # Make sure we got a position
                if math.isnan(x):
                    return

                self.on_pose([x, y, z, rot])

    async def _close(self):
        await self.connection.stream_frames_stop()
        self.connection.disconnect()


def _sqrt(a):
    """
    There might be rounding errors making 'a' slightly negative.
    Make sure we don't throw an exception.
    """
    if a < 0.0:
        return 0.0
    return math.sqrt(a)


def send_extpose_rot_matrix(cf, x, y, z, rot):
    """
    Send the current Crazyflie X, Y, Z position and attitude as a (3x3)
    rotaton matrix. This is going to be forwarded to the Crazyflie's
    position estimator.
    """
    quat = Rotation.from_matrix(rot).as_quat()

    if send_full_pose:
        cf.extpos.send_extpose(x, y, z, quat[0], quat[1], quat[2], quat[3])
    else:
        cf.extpos.send_extpos(x, y, z)


def adjust_orientation_sensitivity(cf):
    cf.param.set_value('locSrv.extQuatStdDev', orientation_std_dev)


def activate_kalman_estimator(cf):
    cf.param.set_value('stabilizer.estimator', '2')

    # Set the std deviation for the quaternion data pushed into the
    # kalman filter. The default value seems to be a bit too low.
    cf.param.set_value('locSrv.extQuatStdDev', 0.06)


def activate_mellinger_controller(cf):
    cf.param.set_value('stabilizer.controller', '2')


def upload_trajectory(cf, trajectory_id, trajectory):
    trajectory_mem = cf.mem.get_mems(MemoryElement.TYPE_TRAJ)[0]
    trajectory_mem.trajectory = []

    total_duration = 0
    for row in trajectory:
        duration = row[0]
        x = Poly4D.Poly(row[1:9])
        y = Poly4D.Poly(row[9:17])
        z = Poly4D.Poly(row[17:25])
        yaw = Poly4D.Poly(row[25:33])
        trajectory_mem.trajectory.append(Poly4D(duration, x, y, z, yaw))
        total_duration += duration

    trajectory_mem.write_data_sync()
    cf.high_level_commander.define_trajectory(trajectory_id, 0, len(trajectory_mem.trajectory))
    return total_duration


def run_sequence(cf, trajectory_id, duration):
    commander = cf.high_level_commander

    commander.takeoff(1.0, 2.0)
    time.sleep(3.0)
    relative = True
    commander.start_trajectory(trajectory_id, 1.0, relative)
    time.sleep(duration)
    commander.land(0.0, 2.0)
    time.sleep(2)
    commander.stop()


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    # Connect to QTM
    qtm_wrapper = QtmWrapper(rigid_body_name)

    with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
        cf = scf.cf
        trajectory_id = 1

        # Set up a callback to handle data from QTM
        qtm_wrapper.on_pose = lambda pose: send_extpose_rot_matrix(
            cf, pose[0], pose[1], pose[2], pose[3])

        adjust_orientation_sensitivity(cf)
        activate_kalman_estimator(cf)
        # activate_mellinger_controller(cf)
        duration = upload_trajectory(cf, trajectory_id, figure8)
        print('The sequence is {:.1f} seconds long'.format(duration))
        reset_estimator(cf)

        # Arm the Crazyflie
        cf.platform.send_arming_request(True)
        time.sleep(1.0)

        run_sequence(cf, trajectory_id, duration)

    qtm_wrapper.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example of a synchronized swarm choreography using the High level
commander.

The swarm takes off and flies a synchronous choreography before landing.
The take-of is relative to the start position but the Goto are absolute.
The sequence contains a list of commands to be executed at each step.

This example is intended to work with any absolute positioning system.
It aims at documenting how to use the High Level Commander together with
the Swarm class to achieve synchronous sequences.
"""
import threading
import time
from collections import namedtuple
from queue import Queue

import cflib.crtp
from cflib.crazyflie.swarm import CachedCfFactory
from cflib.crazyflie.swarm import Swarm

# Time for one step in second
STEP_TIME = 1

# Possible commands, all times are in seconds
Arm = namedtuple('Arm', [])
Takeoff = namedtuple('Takeoff', ['height', 'time'])
Land = namedtuple('Land', ['time'])
Goto = namedtuple('Goto', ['x', 'y', 'z', 'time'])
# RGB [0-255], Intensity [0.0-1.0]
Ring = namedtuple('Ring', ['r', 'g', 'b', 'intensity', 'time'])
# Reserved for the control loop, do not use in sequence
Quit = namedtuple('Quit', [])

uris = [
    'radio://0/10/2M/E7E7E7E701',  # cf_id 0, startup position [-0.5, -0.5]
    'radio://0/10/2M/E7E7E7E702',  # cf_id 1, startup position [ 0, 0]
    'radio://0/10/2M/E7E7E7E703',  # cf_id 3, startup position [0.5, 0.5]
    # Add more URIs if you want more copters in the swarm
    # URIs in a swarm using the same radio must also be on the same channel
]

sequence = [
    # Step, CF_id,  action
    (0,    0,      Arm()),
    (0,    1,      Arm()),
    (0,    2,      Arm()),

    (0,    0,      Takeoff(0.5, 2)),
    (0,    2,      Takeoff(0.5, 2)),

    (1,    1,      Takeoff(1.0, 2)),

    (2,    0,      Goto(-0.5,  -0.5,   0.5, 1)),
    (2,    2,      Goto(0.5,  0.5,   0.5, 1)),

    (3,    1,      Goto(0,  0,   1, 1)),

    (4,    0,      Ring(255, 255, 255, 0.2, 0)),
    (4,    1,      Ring(255, 0, 0, 0.2, 0)),
    (4,    2,      Ring(255, 255, 255, 0.2, 0)),

    (5,    0,      Goto(0.5, -0.5, 0.5, 2)),
    (5,    2,      Goto(-0.5, 0.5, 0.5, 2)),

    (7,    0,      Goto(0.5, 0.5, 0.5, 2)),
    (7,    2,      Goto(-0.5, -0.5, 0.5, 2)),

    (9,    0,      Goto(-0.5, 0.5, 0.5, 2)),
    (9,    2,      Goto(0.5, -0.5, 0.5, 2)),

    (11,   0,      Goto(-0.5, -0.5, 0.5, 2)),
    (11,   2,      Goto(0.5, 0.5, 0.5, 2)),

    (13,    0,      Land(2)),
    (13,    1,      Land(2)),
    (13,    2,      Land(2)),

    (15,    0,      Ring(0, 0, 0, 0, 5)),
    (15,    1,      Ring(0, 0, 0, 0, 5)),
    (15,    2,      Ring(0, 0, 0, 0, 5)),
]


def activate_mellinger_controller(scf, use_mellinger):
    controller = 1
    if use_mellinger:
        controller = 2
    scf.cf.param.set_value('stabilizer.controller', str(controller))


def arm(scf):
    scf.cf.platform.send_arming_request(True)
    time.sleep(1.0)


def set_ring_color(cf, r, g, b, intensity, time):
    cf.param.set_value('ring.fadeTime', str(time))

    r *= intensity
    g *= intensity
    b *= intensity

    color = (int(r) << 16) | (int(g) << 8) | int(b)

    cf.param.set_value('ring.fadeColor', str(color))


def crazyflie_control(scf):
    cf = scf.cf
    control = controlQueues[uris.index(cf.link_uri)]

    activate_mellinger_controller(scf, False)

    commander = scf.cf.high_level_commander

    # Set fade to color effect and reset to Led-ring OFF
    set_ring_color(cf, 0, 0, 0, 0, 0)
    cf.param.set_value('ring.effect', '14')

    while True:
        command = control.get()
        if type(command) is Quit:
            return
        elif type(command) is Arm:
            arm(scf)
        elif type(command) is Takeoff:
            commander.takeoff(command.height, command.time)
        elif type(command) is Land:
            commander.land(0.0, command.time)
        elif type(command) is Goto:
            commander.go_to(command.x, command.y, command.z, 0, command.time)
        elif type(command) is Ring:
            set_ring_color(cf, command.r, command.g, command.b,
                           command.intensity, command.time)
            pass
        else:
            print('Warning! unknown command {} for uri {}'.format(command,
                                                                  cf.uri))


def control_thread():
    pointer = 0
    step = 0
    stop = False

    while not stop:
        print('Step {}:'.format(step))
        while sequence[pointer][0] <= step:
            cf_id = sequence[pointer][1]
            command = sequence[pointer][2]

            print(' - Running: {} on {}'.format(command, cf_id))
            controlQueues[cf_id].put(command)
            pointer += 1

            if pointer >= len(sequence):
                print('Reaching the end of the sequence, stopping!')
                stop = True
                break

        step += 1
        time.sleep(STEP_TIME)

    for ctrl in controlQueues:
        ctrl.put(Quit())


if __name__ == '__main__':
    controlQueues = [Queue() for _ in range(len(uris))]

    cflib.crtp.init_drivers()
    factory = CachedCfFactory(rw_cache='./cache')
    with Swarm(uris, factory=factory) as swarm:
        swarm.reset_estimators()

        print('Starting sequence!')

        threading.Thread(target=control_thread).start()

        swarm.parallel_safe(crazyflie_control)

        time.sleep(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Callback objects used in the Crazyflie library
"""
from threading import Event

__author__ = 'Bitcraze AB'
__all__ = ['Caller']


class Caller():
    """ An object were callbacks can be registered and called """

    def __init__(self):
        """ Create the object """
        self.callbacks = []

    def add_callback(self, cb):
        """ Register cb as a new callback. Will not register duplicates. """
        if ((cb in self.callbacks) is False):
            self.callbacks.append(cb)

    def remove_callback(self, cb):
        """ Un-register cb from the callbacks """
        self.callbacks.remove(cb)

    def call(self, *args):
        """ Call the callbacks registered with the arguments args """
        copy_of_callbacks = list(self.callbacks)
        for cb in copy_of_callbacks:
            cb(*args)


class Syncer:
    """A class to create synchronous behavior for methods using callbacks"""

    def __init__(self):
        self._event = Event()
        self.success_args = None
        self.failure_args = None
        self.is_success = False

    def success_cb(self, *args):
        self.success_args = args
        self.is_success = True
        self._event.set()

    def failure_cb(self, *args):
        self.failure_args = args
        self.is_success = False
        self._event.set()

    def wait(self):
        self._event.wait()

    def clear(self):
        self._event.clear()



#!/usr/bin/env python3
# Demo that makes two Crazyflie take off 30cm above the first controller found
# Using the controller trigger it is then possible to 'grab' the closest
# Crazyflie and to make it move.
import math
import sys
import time

import openvr

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils.reset_estimator import reset_estimator

# URI to the Crazyflie to connect to
uri0 = 'radio://0/80/2M/E7E7E7E701'
uri1 = 'radio://0/80/2M/E7E7E7E702'

print('Opening')
vr = openvr.init(openvr.VRApplication_Other)
print('Opened')

# Find first controller or tracker
controllerId = None
poses = vr.getDeviceToAbsoluteTrackingPose(openvr.TrackingUniverseStanding, 0,
                                           openvr.k_unMaxTrackedDeviceCount)
for i in range(openvr.k_unMaxTrackedDeviceCount):
    if poses[i].bPoseIsValid:
        device_class = vr.getTrackedDeviceClass(i)
        if device_class == openvr.TrackedDeviceClass_Controller or \
           device_class == openvr.TrackedDeviceClass_GenericTracker:
            controllerId = i
            break

if controllerId is None:
    print('Cannot find controller or tracker, exiting')
    sys.exit(1)


def position_callback(timestamp, data, logconf):
    x = data['kalman.stateX']
    y = data['kalman.stateY']
    z = data['kalman.stateZ']
    print('pos: ({}, {}, {})'.format(x, y, z))


def start_position_printing(scf):
    log_conf = LogConfig(name='Position', period_in_ms=500)
    log_conf.add_variable('kalman.stateX', 'float')
    log_conf.add_variable('kalman.stateY', 'float')
    log_conf.add_variable('kalman.stateZ', 'float')

    scf.cf.log.add_config(log_conf)
    log_conf.data_received_cb.add_callback(position_callback)
    log_conf.start()


def vector_subtract(v0, v1):
    return [v0[0] - v1[0], v0[1] - v1[1], v0[2] - v1[2]]


def vector_add(v0, v1):
    return [v0[0] + v1[0], v0[1] + v1[1], v0[2] + v1[2]]


def vector_norm(v0):
    return math.sqrt((v0[0] * v0[0]) + (v0[1] * v0[1]) + (v0[2] * v0[2]))


def run_sequence(scf0, scf1):
    cf0 = scf0.cf
    cf1 = scf1.cf

    poses = vr.getDeviceToAbsoluteTrackingPose(
        openvr.TrackingUniverseStanding, 0, openvr.k_unMaxTrackedDeviceCount)
    controller_pose = poses[controllerId]
    pose = controller_pose.mDeviceToAbsoluteTracking
    setpoints = [[-1 * pose[2][3], -1 * pose[0][3] - 0.5, pose[1][3] + 0.3],
                 [-1 * pose[2][3], -1 * pose[0][3] + 0.5, pose[1][3] + 0.3]]

    closest = 0

    grabbed = False
    grab_controller_start = [0, 0, 0]
    grab_setpoint_start = [0, 0, 0]

    while True:
        poses = vr.getDeviceToAbsoluteTrackingPose(
            openvr.TrackingUniverseStanding, 0,
            openvr.k_unMaxTrackedDeviceCount)
        controller_state = vr.getControllerState(controllerId)[1]

        trigger = ((controller_state.ulButtonPressed & 0x200000000) != 0)

        controller_pose = poses[controllerId]
        pose = controller_pose.mDeviceToAbsoluteTracking

        if not grabbed and trigger:
            print('Grab started')
            grab_controller_start = [-1 * pose[2][3], -1 * pose[0][3], pose[1][3]]

            dist0 = vector_norm(vector_subtract(grab_controller_start,
                                                setpoints[0]))
            dist1 = vector_norm(vector_subtract(grab_controller_start,
                                                setpoints[1]))

            if dist0 < dist1:
                closest = 0
            else:
                closest = 1

            grab_setpoint_start = setpoints[closest]

        if grabbed and not trigger:
            print('Grab ended')

        grabbed = trigger

        if trigger:
            curr = [-1 * pose[2][3], -1 * pose[0][3], pose[1][3]]
            setpoints[closest] = vector_add(
                grab_setpoint_start, vector_subtract(curr,
                                                     grab_controller_start))

        cf0.commander.send_position_setpoint(setpoints[0][0],
                                             setpoints[0][1],
                                             setpoints[0][2],
                                             0)
        cf1.commander.send_position_setpoint(setpoints[1][0],
                                             setpoints[1][1],
                                             setpoints[1][2],
                                             0)

        time.sleep(0.02)

    cf0.commander.send_setpoint(0, 0, 0, 0)
    cf1.commander.send_setpoint(0, 0, 0, 0)
    # Make sure that the last packet leaves before the link is closed
    # since the message queue is not flushed before closing
    time.sleep(0.1)


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    with SyncCrazyflie(uri0, cf=Crazyflie(rw_cache='./cache')) as scf0:
        reset_estimator(scf0)
        with SyncCrazyflie(uri1, cf=Crazyflie(rw_cache='./cache')) as scf1:
            reset_estimator(scf1)

            # Arm the Crazyflies
            scf0.cf.platform.send_arming_request(True)
            scf1.cf.platform.send_arming_request(True)
            time.sleep(1.0)

            run_sequence(scf0, scf1)

    openvr.shutdown()



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2018 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
This script shows the basic use of the PositionHlCommander class.

Simple example that connects to the crazyflie at `URI` and runs a
sequence. This script requires some kind of location system.

The PositionHlCommander uses position setpoints.

Change the URI variable to your Crazyflie configuration.
"""
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.positioning.position_hl_commander import PositionHlCommander
from cflib.utils import uri_helper

# URI to the Crazyflie to connect to
uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')


def slightly_more_complex_usage():
    with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
        # Arm the Crazyflie
        scf.cf.platform.send_arming_request(True)
        time.sleep(1.0)

        with PositionHlCommander(
                scf,
                x=0.0, y=0.0, z=0.0,
                default_velocity=0.3,
                default_height=0.5,
                controller=PositionHlCommander.CONTROLLER_PID) as pc:
            # Go to a coordinate
            pc.go_to(1.0, 1.0, 1.0)

            # Move relative to the current position
            pc.right(1.0)

            # Go to a coordinate and use default height
            pc.go_to(0.0, 0.0)

            # Go slowly to a coordinate
            pc.go_to(1.0, 1.0, velocity=0.2)

            # Set new default velocity and height
            pc.set_default_velocity(0.3)
            pc.set_default_height(1.0)
            pc.go_to(0.0, 0.0)


def land_on_elevated_surface():
    with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
        # Arm the Crazyflie
        scf.cf.platform.send_arming_request(True)
        time.sleep(1.0)

        with PositionHlCommander(scf,
                                 default_height=0.5,
                                 default_velocity=0.2,
                                 default_landing_height=0.35,
                                 controller=PositionHlCommander.CONTROLLER_PID) as pc:
            # fly onto a landing platform at non-zero height (ex: from floor to desk, etc)
            pc.forward(1.0)
            pc.left(1.0)
            # land() will be called on context exit, gradually lowering to default_landing_height, then stopping motors


def simple_sequence():
    with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
        # Arm the Crazyflie
        scf.cf.platform.send_arming_request(True)
        time.sleep(1.0)

        with PositionHlCommander(scf, controller=PositionHlCommander.CONTROLLER_PID) as pc:
            pc.forward(1.0)
            pc.left(1.0)
            pc.back(1.0)
            pc.go_to(0.0, 0.0, 1.0)


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    simple_sequence()
    # slightly_more_complex_usage()
    # land_on_elevated_surface()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
The Crazyflie Micro Quadcopter library API used to communicate with the
Crazyflie Micro Quadcopter via a communication link.

The API takes care of scanning, opening and closing the communication link
as well as sending/receiving data from the Crazyflie.

A link is described using an URI of the following format:
    <interface>://<interface defined data>.
See each link for the data that can be included in the URI for that interface.

The two main uses-cases are scanning for Crazyflies available on a
communication link and opening a communication link to a Crazyflie.

Example of scanning for available Crazyflies on all communication links:
```python
cflib.crtp.init_drivers()
available = cflib.crtp.scan_interfaces()
for i in available:
    print "Found Crazyflie on URI [%s] with comment [%s]"
            % (available[0], available[1])
```

Example of connecting to a Crazyflie with known URI (radio dongle 0 and
radio channel 125):
```python
cf = Crazyflie()
cf.open_link("radio://0/125")
...
cf.close_link()
```
"""
import warnings

warnings.simplefilter('always', DeprecationWarning)  # Enbable DeprecationWarnings

__pdoc__ = {}
__pdoc__['cflib.crtp.cflinkcppdriver'] = False
__pdoc__['cflib.cpx.transports'] = False



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
A script to fly 5 Crazyflies in formation. One stays in the center and the
other four fly around it in a circle. Mainly intended to be used with the
Flow deck.
The starting positions are vital and should be oriented like this

     >

^    +    v

     <

The distance from the center to the perimeter of the circle is around 0.5 m

"""
import math
import time

import cflib.crtp
from cflib.crazyflie.swarm import CachedCfFactory
from cflib.crazyflie.swarm import Swarm

# Change uris according to your setup
# URIs in a swarm using the same radio must also be on the same channel
URI0 = 'radio://0/70/2M/E7E7E7E7E7'
URI1 = 'radio://0/110/2M/E7E7E7E702'
URI2 = 'radio://0/94/2M/E7E7E7E7E7'
URI3 = 'radio://0/5/2M/E7E7E7E702'
URI4 = 'radio://0/110/2M/E7E7E7E703'

# d: diameter of circle
# z: altitude
params0 = {'d': 1.0, 'z': 0.3}
params1 = {'d': 1.0, 'z': 0.3}
params2 = {'d': 0.0, 'z': 0.5}
params3 = {'d': 1.0, 'z': 0.3}
params4 = {'d': 1.0, 'z': 0.3}


uris = {
    URI0,
    URI1,
    URI2,
    URI3,
    URI4,
}

params = {
    URI0: [params0],
    URI1: [params1],
    URI2: [params2],
    URI3: [params3],
    URI4: [params4],
}


def poshold(cf, t, z):
    steps = t * 10

    for r in range(steps):
        cf.commander.send_hover_setpoint(0, 0, 0, z)
        time.sleep(0.1)


def run_sequence(scf, params):
    cf = scf.cf

    # Arm the Crazyflie
    cf.platform.send_arming_request(True)
    time.sleep(1.0)

    # Number of setpoints sent per second
    fs = 4
    fsi = 1.0 / fs

    # Compensation for unknown error :-(
    comp = 1.3

    # Base altitude in meters
    base = 0.15

    d = params['d']
    z = params['z']

    poshold(cf, 2, base)

    ramp = fs * 2
    for r in range(ramp):
        cf.commander.send_hover_setpoint(0, 0, 0, base + r * (z - base) / ramp)
        time.sleep(fsi)

    poshold(cf, 2, z)

    for _ in range(2):
        # The time for one revolution
        circle_time = 8

        steps = circle_time * fs
        for _ in range(steps):
            cf.commander.send_hover_setpoint(d * comp * math.pi / circle_time,
                                             0, 360.0 / circle_time, z)
            time.sleep(fsi)

    poshold(cf, 2, z)

    for r in range(ramp):
        cf.commander.send_hover_setpoint(0, 0, 0,
                                         base + (ramp - r) * (z - base) / ramp)
        time.sleep(fsi)

    poshold(cf, 1, base)

    cf.commander.send_stop_setpoint()
    # Hand control over to the high level commander to avoid timeout and locking of the Crazyflie
    cf.commander.send_notify_setpoint_stop()


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    factory = CachedCfFactory(rw_cache='./cache')
    with Swarm(uris, factory=factory) as swarm:
        swarm.reset_estimators()
        swarm.parallel(run_sequence, args_dict=params)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Crazyflie console is used to receive characters printed using printf
from the firmware.
"""
from cflib.crtp.crtpstack import CRTPPort
from cflib.utils.callbacks import Caller

__author__ = 'Bitcraze AB'
__all__ = ['Console']


class Console:
    """
    Crazyflie console is used to receive characters printed using printf
    from the firmware.
    """

    def __init__(self, crazyflie):
        """
        Initialize the console and register it to receive data from the copter.
        """

        self.receivedChar = Caller()
        """
        This member variable is used to setup a callback that will be called
        when text is received from the CONSOLE port of CRTP (0).

        Example:
        ```python
        [...]

        def log_console(self, text):
            self.log_file.write(text)

        [...]

        self.cf.console.receivedChar.add_callback(self.log_console)
        ```
        """

        self.cf = crazyflie
        self.cf.add_port_callback(CRTPPort.CONSOLE, self._incoming)

    def _incoming(self, packet):
        """
        Callback for data received from the copter.
        """
        # This might be done prettier ;-)
        console_text = packet.data.decode('UTF-8')

        self.receivedChar.call(console_text)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
""" CRTP UDP Driver. Work either with the UDP server or with an UDP device
See udpserver.py for the protocol"""
import queue
import re
import socket
import struct
from urllib.parse import urlparse

from .crtpdriver import CRTPDriver
from .crtpstack import CRTPPacket
from .exceptions import WrongUriType

__author__ = 'Bitcraze AB'
__all__ = ['UdpDriver']


class UdpDriver(CRTPDriver):

    def __init__(self):
        None

    def connect(self, uri, linkQualityCallback, linkErrorCallback):
        if not re.search('^udp://', uri):
            raise WrongUriType('Not an UDP URI')

        parse = urlparse(uri)

        self.queue = queue.Queue()
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.addr = (parse.hostname, parse.port)
        self.socket.connect(self.addr)

        self.socket.sendto('\xFF\x01\x01\x01'.encode(), self.addr)

    def receive_packet(self, time=0):
        data, addr = self.socket.recvfrom(1024)

        if data:
            data = struct.unpack('B' * (len(data) - 1), data[0:len(data) - 1])
            pk = CRTPPacket()
            pk.port = data[0]
            pk.data = data[1:]
            return pk

        try:
            if time == 0:
                return self.rxqueue.get(False)
            elif time < 0:
                while True:
                    return self.rxqueue.get(True, 10)
            else:
                return self.rxqueue.get(True, time)
        except queue.Empty:
            return None

    def send_packet(self, pk):
        raw = (pk.port,) + struct.unpack('B' * len(pk.data), pk.data)

        cksum = 0
        for i in raw:
            cksum += i

        cksum %= 256

        data = ''.join(chr(v) for v in (raw + (cksum,)))

        # print tuple(data)
        self.socket.sendto(data.encode(), self.addr)

    def close(self):
        # Remove this from the server clients list
        self.socket.sendto('\xFF\x01\x02\x02'.encode(), self.addr)

    def get_name(self):
        return 'udp'

    def scan_interface(self, address):
        return []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2022 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
""" CPX Router and discovery"""
import enum
import logging
import queue
import struct
import threading

__author__ = 'Bitcraze AB'
__all__ = ['CPXRouter']
logger = logging.getLogger(__name__)


class CPXTarget(enum.Enum):
    """
    List of CPX targets
    """
    STM32 = 1
    ESP32 = 2
    HOST = 3
    GAP8 = 4


class CPXFunction(enum.Enum):
    """
    List of CPX targets
    """
    SYSTEM = 1
    CONSOLE = 2
    CRTP = 3
    WIFI_CTRL = 4
    APP = 5
    TEST = 0x0E
    BOOTLOADER = 0x0F


class CPXPacket(object):
    """
    A packet with routing and data
    """
    CPX_VERSION = 0

    def __init__(self, function=0, destination=0, source=CPXTarget.HOST, data=bytearray()):
        """
        Create an empty packet with default values.
        """
        self.data = data
        self.source = source
        self.destination = destination
        self.function = function
        self.version = self.CPX_VERSION
        self.length = len(data)
        self.lastPacket = False

    def _get_wire_data(self):
        """Create raw data to send via the wire"""
        raw = bytearray()

        targetsAndFlags = ((self.source.value & 0x7) << 3) | (self.destination.value & 0x7)
        if self.lastPacket:
            targetsAndFlags |= 0x40

        functionAndVersion = (self.function.value & 0x3F) | ((self.version & 0x3) << 6)
        raw.extend(struct.pack('<BB', targetsAndFlags, functionAndVersion))
        raw.extend(self.data)

        return raw

    def _set_wire_data(self, data):
        [targetsAndFlags, functionAndVersion] = struct.unpack('<BB', data[0:2])
        self.version = (functionAndVersion >> 6) & 0x3
        if self.version != self.CPX_VERSION:
            logging.error(f'Unsupported CPX version {self.version} instead of {self.CPX_VERSION}')
            raise RuntimeError(f'Unsupported CPX version {self.version} instead of {self.CPX_VERSION}')
        self.source = CPXTarget((targetsAndFlags >> 3) & 0x07)
        self.destination = CPXTarget(targetsAndFlags & 0x07)
        self.lastPacket = targetsAndFlags & 0x40 != 0
        self.function = CPXFunction(functionAndVersion & 0x3F)
        self.data = data[2:]
        self.length = len(self.data)

    def __str__(self):
        """Get a string representation of the packet"""
        return '{}->{}/{} (size={} bytes)'.format(self.source, self.destination, self.function, self.length)

    wireData = property(_get_wire_data, _set_wire_data)


class CPXRouter(threading.Thread):

    def __init__(self, transport):
        threading.Thread.__init__(self)
        self.daemon = True
        self._transport = transport
        self._rxQueues = {}
        self._packet_assembly = []
        self._connected = True

    # Register and/or blocking calls for ports
    def receivePacket(self, function, timeout=None):

        # Check if a queue exists, if not then create it
        # the user might have implemented new functions
        if function.value not in self._rxQueues:
            print('Creating queue for {}'.format(function))
            self._rxQueues[function.value] = queue.Queue()

        return self._rxQueues[function.value].get(block=True, timeout=timeout)

    def makeTransaction(self, packet):
        self.sendPacket(packet)
        return self.receivePacket(packet.function)

    def sendPacket(self, packet):
        self._transport.writePacket(packet)

    def transport(self):
        self._connected = False
        return self._transport

    def run(self):
        while (self._connected):
            # Read one packet from the transport

            # Packages might have been split up along the
            # way, we should re-assemble here
            try:
                packet = self._transport.readPacket()

                if packet.function.value not in self._rxQueues:
                    pass
                else:
                    self._rxQueues[packet.function.value].put(packet)
            except Exception as e:
                print('Exception while reading transport, link probably closed?')
                print(e)
                import traceback
                print(traceback.format_exc())

# Public facing


class CPX:
    def __init__(self, transport):
        self._router = CPXRouter(transport)
        self._router.start()

    def receivePacket(self, function, timeout=None):
        # Block on a function queue

        # if self._router.isUsedBySubmodule(target, function):
        #  raise ValueError("The CPX target {} and function {} is registered in a sub module".format(target, function))

        # Will block on queue
        return self._router.receivePacket(function, timeout)

    def makeTransaction(self, packet):
        return self._router.makeTransaction(packet)

    def sendPacket(self, packet):
        self._router.sendPacket(packet)

    def close(self):
        self._router.transport().disconnect()



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import logging
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.crazyflie.syncLogger import SyncLogger
from cflib.utils import uri_helper

# URI to the Crazyflie to connect to
uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


def param_stab_est_callback(name, value):
    print('The crazyflie has parameter ' + name + ' set at number: ' + value)


def simple_param_async(scf, groupstr, namestr):
    cf = scf.cf
    full_name = groupstr + '.' + namestr

    cf.param.add_update_callback(group=groupstr, name=namestr,
                                 cb=param_stab_est_callback)
    time.sleep(1)
    cf.param.set_value(full_name, 2)
    time.sleep(1)
    cf.param.set_value(full_name, 1)
    time.sleep(1)


def log_stab_callback(timestamp, data, logconf):
    print('[%d][%s]: %s' % (timestamp, logconf.name, data))


def simple_log_async(scf, logconf):
    cf = scf.cf
    cf.log.add_config(logconf)
    logconf.data_received_cb.add_callback(log_stab_callback)
    logconf.start()
    time.sleep(5)
    logconf.stop()


def simple_log(scf, logconf):

    with SyncLogger(scf, logconf) as logger:

        for log_entry in logger:

            timestamp = log_entry[0]
            data = log_entry[1]
            logconf_name = log_entry[2]

            print('[%d][%s]: %s' % (timestamp, logconf_name, data))

            break


def simple_connect():

    print("Yeah, I'm connected! :D")
    time.sleep(3)
    print("Now I will disconnect :'(")


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    lg_stab = LogConfig(name='Stabilizer', period_in_ms=10)
    lg_stab.add_variable('stabilizer.roll', 'float')
    lg_stab.add_variable('stabilizer.pitch', 'float')
    lg_stab.add_variable('stabilizer.yaw', 'float')

    group = 'stabilizer'
    name = 'estimator'

    with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:

        # simple_connect()

        # simple_log(scf, lg_stab)

        # simple_log_async(scf, lg_stab)

        simple_param_async(scf, group, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2018-2020 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Used for sending high level setpoints to the Crazyflie.

The high level commander generates setpoints from within the firmware
based on a predefined trajectory. This was merged as part of the
Crazyswarm project of the USC ACT lab. The high level commander uses a
planner to generate smooth trajectories based on actions like "take off",
"go to" or "land" with 7th order polynomials.
"""
import math
import struct

from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort

__author__ = 'Bitcraze AB'
__all__ = ['HighLevelCommander']


class HighLevelCommander():
    """
    Used for sending high level setpoints to the Crazyflie
    """

    COMMAND_SET_GROUP_MASK = 0
    COMMAND_STOP = 3
    COMMAND_GO_TO = 4
    COMMAND_START_TRAJECTORY = 5
    COMMAND_DEFINE_TRAJECTORY = 6
    COMMAND_TAKEOFF_2 = 7
    COMMAND_LAND_2 = 8
    COMMAND_SPIRAL = 11
    COMMAND_GO_TO_2 = 12
    COMMAND_START_TRAJECTORY_2 = 13

    ALL_GROUPS = 0

    TRAJECTORY_LOCATION_MEM = 1

    TRAJECTORY_TYPE_POLY4D = 0
    TRAJECTORY_TYPE_POLY4D_COMPRESSED = 1

    def __init__(self, crazyflie=None):
        """
        Initialize the object.
        """
        self._cf = crazyflie

    def set_group_mask(self, group_mask=ALL_GROUPS):
        """
        Set the group mask that the Crazyflie belongs to

        :param group_mask: Mask for which groups this CF belongs to
        """
        self._send_packet(struct.pack('<BB',
                                      self.COMMAND_SET_GROUP_MASK,
                                      group_mask))

    def takeoff(self, absolute_height_m, duration_s, group_mask=ALL_GROUPS,
                yaw=0.0):
        """
        vertical takeoff from current x-y position to given height

        :param absolute_height_m: Absolute (m)
        :param duration_s: Time it should take until target height is
                           reached (s)
        :param group_mask: Mask for which CFs this should apply to
        :param yaw: Yaw (rad). Use current yaw if set to None.
        """
        target_yaw = yaw
        useCurrentYaw = False
        if yaw is None:
            target_yaw = 0.0
            useCurrentYaw = True

        self._send_packet(struct.pack('<BBff?f',
                                      self.COMMAND_TAKEOFF_2,
                                      group_mask,
                                      absolute_height_m,
                                      target_yaw,
                                      useCurrentYaw,
                                      duration_s))

    def land(self, absolute_height_m, duration_s, group_mask=ALL_GROUPS,
             yaw=0.0):
        """
        vertical land from current x-y position to given height

        :param absolute_height_m: Absolute (m)
        :param duration_s: Time it should take until target height is
                           reached (s)
        :param group_mask: Mask for which CFs this should apply to
        :param yaw: Yaw (rad). Use current yaw if set to None.
        """
        target_yaw = yaw
        useCurrentYaw = False
        if yaw is None:
            target_yaw = 0.0
            useCurrentYaw = True

        self._send_packet(struct.pack('<BBff?f',
                                      self.COMMAND_LAND_2,
                                      group_mask,
                                      absolute_height_m,
                                      target_yaw,
                                      useCurrentYaw,
                                      duration_s))

    def stop(self, group_mask=ALL_GROUPS):
        """
        stops the current trajectory (turns off the motors)

        :param group_mask: Mask for which CFs this should apply to
        :return:
        """
        self._send_packet(struct.pack('<BB',
                                      self.COMMAND_STOP,
                                      group_mask))

    def go_to(self, x, y, z, yaw, duration_s, relative=False, linear=False,
              group_mask=ALL_GROUPS):
        """
        Go to an absolute or relative position.

        The path is designed to transition smoothly from the current
        state to the target position, gradually decelerating at the
        goal with minimal overshoot. When the system is at hover, the
        path will be a straight line, but if there is any initial
        velocity, the path will be a smooth curve.

        The trajectory is derived by solving for a unique 7th-degree
        polynomial that satisfies the initial conditions of position,
        velocity, and acceleration, and ends at the goal with zero
        velocity and acceleration. Additionally, the jerk (derivative
        of acceleration) is constrained to be zero at both the starting
        and ending points.

        Warning! Avoid overlapping go_to commands. When a command is sent to a
        Crazyflie when another one is currently executed, the generated polynomial
        can take unexpected routes and have high peaks.

        :param x: X (m)
        :param y: Y (m)
        :param z: Z (m)
        :param yaw: Yaw (radians)
        :param duration_s: Time it should take to reach the position (s)
        :param relative: True if x, y, z is relative to the current position
        :param linear: True to use linear interpolation instead of a smooth polynomial
        :param group_mask: Mask for which CFs this should apply to
        """
        if self._cf.platform.get_protocol_version() < 8:
            if linear:
                print('Warning: Linear mode not supported in protocol version < 8, update your crazyflie\'s firmware')
            self._send_packet(struct.pack('<BBBfffff',
                                          self.COMMAND_GO_TO,
                                          group_mask,
                                          relative,
                                          x, y, z,
                                          yaw,
                                          duration_s))
        else:
            self._send_packet(struct.pack('<BBBBfffff',
                                          self.COMMAND_GO_TO_2,
                                          group_mask,
                                          relative,
                                          linear,
                                          x, y, z,
                                          yaw,
                                          duration_s))

    def spiral(self, angle, r0, rF, ascent, duration_s, sideways=False, clockwise=False,
               group_mask=ALL_GROUPS):
        """
        Follow a spiral-like segment (spline approximation of a spiral/arc for <= 90-degree segments)

        :param angle: spiral angle (rad), limited to +/- 2pi
        :param r0: initial radius (m), must be positive
        :param rF: final radius (m), must be positive
        :param ascent: altitude gain (m), positive to climb, negative to descent
        :param duration_s: time it should take to reach the end of the spiral (s)
        :param sideways: true if crazyflie should spiral sideways instead of forward
        :param clockwise: true if crazyflie should spiral clockwise instead of counter-clockwise
        :param group_mask: Mask for which CFs this should apply to
        """
        if self._cf.platform.get_protocol_version() < 8:
            print('Warning: Spiral command is not supported in protocol version < 8, update your crazyflie\'s firmware')
        else:
            if angle > 2*math.pi:
                angle = 2*math.pi
                print('Warning: Spiral angle saturated at 2pi as it was too large')
            elif angle < -2*math.pi:
                angle = -2*math.pi
                print('Warning: Spiral angle saturated at -2pi as it was too small')
            if r0 < 0:
                r0 = 0
                print('Warning: Initial radius set to 0 as it cannot be negative')
            if rF < 0:
                rF = 0
                print('Warning: Final radius set to 0 as it cannot be negative')
            self._send_packet(struct.pack('<BBBBfffff',
                                          self.COMMAND_SPIRAL,
                                          group_mask,
                                          sideways,
                                          clockwise,
                                          angle,
                                          r0, rF,
                                          ascent,
                                          duration_s))

    def start_trajectory(self, trajectory_id, time_scale=1.0, relative_position=False,
                         relative_yaw=False, reversed=False, group_mask=ALL_GROUPS):
        """
        starts executing a specified trajectory

        :param trajectory_id: Id of the trajectory (previously defined by
               define_trajectory)
        :param time_scale: Time factor; 1.0 = original speed;
                                        >1.0: slower;
                                        <1.0: faster
        :param relative_position: Set to True, if trajectory should be shifted to
               current setpoint
        :param relative_yaw: Set to True, if trajectory should be aligned to
               current yaw
        :param reversed: Set to True, if trajectory should be executed in
               reverse
        :param group_mask: Mask for which CFs this should apply to
        :return:
        """
        if self._cf.platform.get_protocol_version() < 10:
            if relative_yaw:
                print('Warning: Relative yaw is not supported in protocol '
                      'version < 9, update your Crazyflie\'s firmware')
            self._send_packet(struct.pack('<BBBBBf',
                                          self.COMMAND_START_TRAJECTORY,
                                          group_mask,
                                          relative_position,
                                          reversed,
                                          trajectory_id,
                                          time_scale))
        else:
            self._send_packet(struct.pack('<BBBBBBf',
                                          self.COMMAND_START_TRAJECTORY_2,
                                          group_mask,
                                          relative_position,
                                          relative_yaw,
                                          reversed,
                                          trajectory_id,
                                          time_scale))

    def define_trajectory(self, trajectory_id, offset, n_pieces, type=TRAJECTORY_TYPE_POLY4D):
        """
        Define a trajectory that has previously been uploaded to memory.

        :param trajectory_id: The id of the trajectory
        :param offset: Offset in uploaded memory
        :param n_pieces: Nr of pieces in the trajectory
        :param type: The type of trajectory data; TRAJECTORY_TYPE_POLY4D or TRAJECTORY_TYPE_POLY4D_COMPRESSED
        :return:
        """
        self._send_packet(struct.pack('<BBBBIB',
                                      self.COMMAND_DEFINE_TRAJECTORY,
                                      trajectory_id,
                                      self.TRAJECTORY_LOCATION_MEM,
                                      type,
                                      offset,
                                      n_pieces))

    def _send_packet(self, data):
        pk = CRTPPacket()
        pk.port = CRTPPort.SETPOINT_HL
        pk.data = data
        self._cf.send_packet(pk)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2021-2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from cflib.localization import LighthouseBsVector
from cflib.localization.lighthouse_bs_vector import LighthouseBsVectors


class LighthouseSweepAngleReader():
    """
    Wrapper to simplify reading of lighthouse sweep angles from the locSrv stream
    """
    ANGLE_STREAM_PARAM = 'locSrv.enLhAngleStream'
    NR_OF_SENSORS = 4

    def __init__(self, cf, data_recevied_cb):
        self._cf = cf
        self._cb = data_recevied_cb
        self._is_active = False

    def start(self):
        """Start reading sweep angles"""
        self._cf.loc.receivedLocationPacket.add_callback(self._packet_received_cb)
        self._angle_stream_activate(True)
        self._is_active = True

    def stop(self):
        """Stop reading sweep angles"""
        if self._is_active:
            self._is_active = False
            self._cf.loc.receivedLocationPacket.remove_callback(self._packet_received_cb)
            self._angle_stream_activate(False)

    def _angle_stream_activate(self, is_active):
        value = 0
        if is_active:
            value = 1
        self._cf.param.set_value(self.ANGLE_STREAM_PARAM, value)

    def _packet_received_cb(self, packet):
        if packet.type != self._cf.loc.LH_ANGLE_STREAM:
            return

        if self._cb:
            base_station_id = packet.data['basestation']
            horiz_angles = packet.data['x']
            vert_angles = packet.data['y']

            result = []
            for i in range(self.NR_OF_SENSORS):
                result.append(LighthouseBsVector(horiz_angles[i], vert_angles[i]))

            self._cb(base_station_id, LighthouseBsVectors(result))


class LighthouseSweepAngleAverageReader():
    """
    Helper class to make it easy read sweep angles for multiple base stations and average the result
    """

    def __init__(self, cf, ready_cb):
        self._reader = LighthouseSweepAngleReader(cf, self._data_recevied_cb)
        self._ready_cb = ready_cb
        self.nr_of_samples_required = 50

        # We store all samples in the storage for averaging when data is collected
        # The storage is a dictionary keyed on the base station channel
        # Each entry is a list of 4 lists, one per sensor.
        # Each list contains LighthouseBsVector objects, representing the sampled sweep angles
        self._sample_storage = None

    def start_angle_collection(self):
        """
        Start collecting angles. The process will terminate when nr_of_samples_required have been
        received
        """
        self._sample_storage = {}
        self._reader.start()

    def stop_angle_collection(self):
        """Premature stop of data collection"""
        self._reader.stop()
        self._sample_storage = None

    def is_collecting(self):
        """True if data collection is in progress"""
        return self._sample_storage is not None

    def _data_recevied_cb(self, base_station_id, bs_vectors):
        self._store_sample(base_station_id, bs_vectors, self._sample_storage)
        if self._has_collected_enough_data(self._sample_storage):
            self._reader.stop()
            if self._ready_cb:
                averages = self._average_all_lists(self._sample_storage)
                self._ready_cb(averages)
            self._sample_storage = None

    def _store_sample(self, base_station_id, bs_vectors, storage):
        if base_station_id not in storage:
            storage[base_station_id] = []
            for sensor in range(self._reader.NR_OF_SENSORS):
                storage[base_station_id].append([])

        for sensor in range(self._reader.NR_OF_SENSORS):
            storage[base_station_id][sensor].append(bs_vectors[sensor])

    def _has_collected_enough_data(self, storage):
        for sample_list in storage.values():
            if len(sample_list[0]) >= self.nr_of_samples_required:
                return True
        return False

    def _average_all_lists(self, storage):
        result = {}

        for id, sample_lists in storage.items():
            averages = self._average_sample_lists(sample_lists)
            count = len(sample_lists[0])
            result[id] = (count, averages)

        return result

    def _average_sample_lists(self, sample_lists):
        result = []

        for i in range(self._reader.NR_OF_SENSORS):
            result.append(self._average_sample_list(sample_lists[i]))

        return LighthouseBsVectors(result)

    def _average_sample_list(self, sample_list):
        sum_horiz = 0.0
        sum_vert = 0.0

        for bs_vector in sample_list:
            sum_horiz += bs_vector.lh_v1_horiz_angle
            sum_vert += bs_vector.lh_v1_vert_angle

        count = len(sample_list)
        return LighthouseBsVector(sum_horiz / count, sum_vert / count)
